'''
Since some covid preprint papers could not be processed with papers class (in papers/raw.py), many methods are overwritten in
paper_covid_preprint and papers_covid_preprint.

[Documentation here can be more detailed if needed]
'''

import ast
from tqdm.notebook import tqdm
from naimai.papers.raw import papers, paper_base
import pandas as pd
from naimai.utils.regex import replace_abbreviations

class paper_covid_preprint(paper_base):
    def __init__(self, data,obj_classifier_model=None):
        super().__init__(obj_classifier_model)
        self.database = 'covid preprint - '
        self.data = data

    def get_Abstract(self):
        self.Abstract = self.data['Abstract']

    def get_Title(self):
        self.Title = self.data['Title of preprint']

    def get_doi(self):
        self.doi = self.data['DOI']

    def get_Authors(self):
        self.Authors = ', '.join(ast.literal_eval(self.data['Authors']))

    def update_database(self):
        self.database = self.database + self.data['Uploaded Site']

    def get_Publication_year(self):
        self.Publication_year = self.data['Date of Upload'].split('-')[0]


class papers_covid_preprint(papers):

    def __init__(self, obj_classifier_model=None, author_classifier_model=None,
                 load_obj_classifier_model=True, load_author_classifier_model=False, load_nlp=True):
        super().__init__(obj_classifier_model=obj_classifier_model, author_classifier_model=author_classifier_model,
                         load_obj_classifier_model=load_obj_classifier_model,
                         load_author_classifier_model=load_author_classifier_model,
                         load_nlp=load_nlp)
        self.data = pd.read_csv('drive/MyDrive/MyProject/data/COVID-19-Preprint-Data_ver5.csv')

    def add_paper(self, idx, save_dict=True, report=True):
        data = self.data.iloc[idx, :]
        new_paper = paper_covid_preprint(data=data)
        new_paper.file_name = idx
        new_paper.get_Abstract()
        new_paper.get_Title()
        new_paper.get_doi()
        new_paper.get_Authors()
        new_paper.update_database()
        new_paper.get_Publication_year()
        new_paper = replace_abbreviations(new_paper)
        new_paper.get_objective_paper()
        if report:
          new_paper.report_objectives()
        if save_dict:
            self.elements[new_paper.file_name] = new_paper.save_paper_for_training()
            self.naimai_elements[new_paper.file_name] = new_paper.save_paper_for_naimai()
        else:
            self.elements[new_paper.file_name] = new_paper

    def get_papers(self, save_dict=True, report=True):
        for idx in tqdm(range(len(self.data))):
            self.add_paper(idx=idx, save_dict=save_dict, report=report)



import numpy as np
import pandas as  pd
from sklearn.metrics.pairwise import linear_kernel
from sklearn.feature_extraction.text import TfidfVectorizer

class tfidf_model:
    def __init__(self, query: str, papers: dict, vectorizer=None):
        self.papers = papers
        self.papers_fnames = ['query'] + [fname for fname in papers]
        self.query = query
        self.corpus = []
        self.encodings = None
        if vectorizer:
            self.vectorizer = vectorizer
        else:
            self.vectorizer = TfidfVectorizer(use_idf=True, stop_words="english")

    def paper2doc(self, paper: dict) -> str:
        '''
        transform paper dict to a document : text from its messages
        '''
        messages = ' '.join(paper['messages'])
        return messages

    def build_corpus(self) -> list:
        '''
        build corpus based on the messages of papers : list of texts from messages of papers
        '''
        corpus = [self.query, ]
        for fname in self.papers:
            paper = self.papers[fname]
            doc = self.paper2doc(paper)
            corpus.append(doc)
        self.corpus = corpus

    def get_tf_encodings(self):
        '''
        get tf encodings by computing tfidf scores
        '''
        encodings = self.vectorizer.fit_transform(self.corpus).toarray()
        encodings_df = pd.DataFrame(encodings, columns=self.vectorizer.get_feature_names_out(), index=self.papers_fnames)
        df = encodings_df.T[(encodings_df.T > 0).any(1)]
        self.encodings = df.T

    def most_similar(self, top_n=5) -> tuple:
        '''
        get fnames of similar papers using encodings
        '''
        cosine_scores = linear_kernel(self.encodings, self.encodings)
        cos_score = np.array(cosine_scores[0])
        most_sims = np.argsort(cos_score)[::-1][1:(top_n + 1)]
        similar_fnames = [self.papers_fnames[idx] for idx in most_sims]
        cos_scores2 = [cos_score[i] for i in most_sims]
        # result = [(message, cos_score[i]) for message, i in zip(similar_messages, most_sims) if cos_score[i]>0]
        return similar_fnames, cos_scores2

    def get_similar_fnames(self, top_n=20) -> tuple:
        ''' get similar fnames '''
        self.build_corpus()
        self.get_tf_encodings()
        fnames = self.most_similar(top_n=top_n)
        return fnames


import torch
from sklearn.metrics import classification_report
from spacy import displacy
from naimai.constants.models import colors_labels
import numpy as np
from naimai.constants.models import output_labels
from datasets import load_metric
import pandas as pd

metric = load_metric("seqeval")

def compute_metrics(eval_pred):
    predictions, labels = eval_pred
    predictions = np.argmax(predictions, axis=2)

    # Remove ignored index (special tokens)
    true_predictions = [
        [output_labels[eval_pred] for (eval_pred, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, labels)
    ]
    true_labels = [
        [output_labels[l] for (eval_pred, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, labels)
    ]

    results = metric.compute(predictions=true_predictions, references=true_labels)
    return {
        "accuracy": results["overall_accuracy"],
    }


def sklearn_scores(labels,predictions):
  scores=classification_report(labels,predictions, output_dict=True, zero_division=0)
  result = {'f1 macro avg': scores['macro avg']['f1-score'],
            'accuracy': scores['accuracy']}
  return result

def get_ids_mask_labels(batch,device,labels_from_data=True, model=None):
  ids = batch['input_ids'].to(device, dtype = torch.long)
  mask = batch['attention_mask'].to(device, dtype = torch.long)
  if labels_from_data:
    labels = batch['labels'].to(device, dtype = torch.long)
  else:
    labels = model(ids, attention_mask=mask, return_dict=False)
  return (ids,mask,labels)

def flatten_labels_and_predictions(model,labels,tr_logits):
  flattened_labels = labels.view(-1) # shape (batch_size * seq_len,)
  active_logits = tr_logits.view(-1, model.num_labels) # shape (batch_size * seq_len, num_labels)
  flattened_predictions = torch.argmax(active_logits, axis=1) # shape (batch_size * seq_len,)
  return (flattened_labels,flattened_predictions)


def calc_overlap(row):
    """
    Calculates the overlap between prediction and
    ground truth and overlap percentages used for determining
    true positives.
    """
    set_pred = set(row.predictionstring_pred.split(' '))
    set_gt = set(row.predictionstring_gt.split(' '))
    # Length of each and intersection
    len_gt = len(set_gt)
    len_pred = len(set_pred)
    inter = len(set_gt.intersection(set_pred))
    overlap_1 = inter / len_gt
    overlap_2 = inter / len_pred
    return [overlap_1, overlap_2]


def score_feedback_comp(pred_df, gt_df):
    """
    Modified function that scores for the kaggle
        Student Writing Competition

    Uses the steps in the evaluation page here:
        https://www.kaggle.com/c/feedback-prize-2021/overview/evaluation
    """
    gt_df = gt_df[['doi', 'class', 'predictionstring']].reset_index(drop=True).copy()
    pred_df = pred_df[['doi', 'class', 'predictionstring']].reset_index(drop=True).copy()
    pred_df['pred_id'] = pred_df.index
    gt_df['gt_id'] = gt_df.index
    # Step 1. all ground truths and predictions for a given class are compared.
    joined = pred_df.merge(gt_df,
                           left_on=['doi', 'class'],
                           right_on=['doi', 'class'],
                           how='outer',
                           suffixes=('_pred', '_gt')
                           )
    joined['predictionstring_gt'] = joined['predictionstring_gt'].fillna(' ')
    joined['predictionstring_pred'] = joined['predictionstring_pred'].fillna(' ')

    joined['overlaps'] = joined.apply(calc_overlap, axis=1)

    # 2. If the overlap between the ground truth and prediction is >= 0.5,
    # and the overlap between the prediction and the ground truth >= 0.5,
    # the prediction is a match and considered a true positive.
    # If multiple matches exist, the match with the highest pair of overlaps is taken.
    joined['overlap1'] = joined['overlaps'].apply(lambda x: eval(str(x))[0])
    joined['overlap2'] = joined['overlaps'].apply(lambda x: eval(str(x))[1])

    joined['potential_TP'] = (joined['overlap1'] >= 0.5) & (joined['overlap2'] >= 0.5)
    joined['max_overlap'] = joined[['overlap1', 'overlap2']].max(axis=1)
    tp_pred_ids = joined.query('potential_TP') \
        .sort_values('max_overlap', ascending=False) \
        .groupby(['doi', 'predictionstring_gt']).first()['pred_id'].values

    # 3. Any unmatched ground truths are false negatives
    # and any unmatched predictions are false positives.
    fp_pred_ids = [p for p in joined['pred_id'].unique() if p not in tp_pred_ids]

    matched_gt_ids = joined.query('potential_TP')['gt_id'].unique()
    unmatched_gt_ids = [c for c in joined['gt_id'].unique() if c not in matched_gt_ids]

    # Get numbers of each type
    TP = len(tp_pred_ids)
    FP = len(fp_pred_ids)
    FN = len(unmatched_gt_ids)
    # calc microf1
    my_f1_score = TP / (TP + 0.5 * (FP + FN))

    return my_f1_score


def split_list(lst) -> list:
    '''
    split list lst into list of sequences retrieved in lst
    :param lst:
    :return:
    '''
    preceding_elt = lst[0]
    new_starting_idxs = []
    for idx, elt in enumerate(lst[1:]):
        if elt == preceding_elt + 1:
            preceding_elt = elt
        else:
            new_starting_idxs.append(idx + 1)
            preceding_elt = elt

    if new_starting_idxs:
        first_elt = lst[0]
        new_list = []
        for idx in new_starting_idxs:
            ls_ = list(np.arange(first_elt, lst[idx - 1] + 1))
            new_list.append(ls_)
            try:
                first_elt = lst[idx + 1] - 1
            except:
                first_elt = lst[idx]
        last_elt = list(np.arange(first_elt, lst[-1] + 1))
        new_list.append(last_elt)

        return new_list
    else:
        return [lst]

def list2pstring(lst):
  # convert list to predictionstring format
  lst_str = list(map(str,lst))
  return ' '.join(lst_str)

def get_first_char_id(elt,text):
  start_wd = elt['start']
  split=text.split()
  first_wds = ' '.join(split[start_wd:start_wd+5])
  first_char = text.index(first_wds)
  return first_char

def get_last_char_id(elt,text):
  end_wid = elt['end']
  split=text.split()
  last_wds = ' '.join(split[end_wid-5:end_wid])
  end_char = text.index(last_wds)+len(last_wds)
  return end_char

def get_doc_options(txt, df):
    '''
    get necessary elements to visualize ner results with displacy
    :param txt:
    :param df:
    :return:
    '''
    labels_list = df['class'].tolist()
    labels_list = [elt[:3].upper() for elt in labels_list]
    start_list = df['start_char'].tolist()
    end_list = df['last_char'].tolist()

    ents = []
    colors = {label: colors_labels[label] for label in labels_list}
    for start, end, label in zip(start_list, end_list, labels_list):
        dic = {'start': start, 'end': end, 'label': label}
        ents.append(dic)

    doc = {'text': txt, "ents": ents}
    options = {'ents': labels_list, "colors": colors}
    return {'doc': doc, "options": options}


def visualize(txt, df,show=True):
    '''
    Visualize segmented text in data frame df using displacy
    :param txt:
    :param df:
    :return:
    '''
    results = get_doc_options(txt, df)
    if show:
        doc = results['doc']
        options = results['options']
        displacy.render(doc, style="ent", options=options, manual=True, jupyter=True)


def get_text(elt,txt):
    '''
    used in df.apply to extract text using start char and last char of predicted classes.
    :param elt:
    :param txt:
    :return:
    '''
    start_char = elt['start_char']
    last_char = elt['last_char']
    return txt[start_char:last_char]



def review_paper_error_log_decorator(func):
  def wrapper(*args,**kwargs):
    try:
      return func(*args,**kwargs)
    except:
      pass
      # error_txt = 'review_paper problem in ' + str(args[1].pdf_path) + '\n'
      # with open(args[0].path_errors_log + 'review_paper_errors_log.txt', 'a') as f:
      #     f.write(error_txt)
  return wrapper

def paper_reading_error_log_decorator(func):
  def wrapper(*args,**kwargs):
    try:
      func(*args,**kwargs)
    except:
      pass
      # error_txt = 'paper reading problem in ' + str(args[1]) + '\n'
      # with open(args[0].path_errors_log + 'paper_reading_errors_log.txt', 'a') as f:
      #     f.write(error_txt)
  return wrapper

def update_naimai_dois(func):
  def wrapper(*args,**kwargs):
    func(*args,**kwargs)
    if kwargs['update_dois']:
      print('>> Updating naimai dois..')
      args[0].update_naimai_dois()
      print('Done !')

  return wrapper


'''
Extract data from an article in PDF using grobid package modified for naimai.
The modified grobid package is in the grobid file.
'''

from naimai.models.abbreviation import extract_abbreviation_definition_pairs
from naimai.constants.regex import regex_paper_year
from naimai.constants.paths import path_open_citations, grobid_url
from naimai.papers.full_text.grobid.pdf2dict import GrobidClient
from naimai.utils.general import get_soup
from naimai.utils.regex import multiple_replace
from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import ast
import os
import re

class paper_pdf:
    def __init__(self,fname,pdf_content):
        self.database = 'pdf'
        self.fname = fname
        self.content = BeautifulSoup(pdf_content, "lxml")
        self.numCitedBy = .5
        self.Journal = ''
        self.Abstract = ''
        self.Authors = ''
        self.doi = ''
        self.year = 999
        self.Title = ''

        # if paper_path:
        #     self.read_xml_file(paper_path, is_path=True)
        # else:
        #     self.read_xml_file(paper_xml, is_path=False)

    # def read_xml_file(self, xml, is_path):
    #     '''
    #     read xml file (converted using grobid)
    #     :param xml:
    #     :param is_path:
    #     :return:
    #     '''
    #     if is_path:
    #         xml_file = open(xml, 'r')
    #     else:
    #         xml_file = xml
    #     self.xml_soup = BeautifulSoup(xml_file, "lxml")


    def get_data(self,name,attrs={},findall=False,is_text=True):
        '''
        get data using name and attrs
        :param name:
        :param attrs:
        :param findall:
        :return:
        '''
        if findall:
            # text = self.xml_soup.find_all(name=name, attrs=attrs)
            text = self.content.find_all(name=name, attrs=attrs)
            if text:
              return text[0]
        else:
            # text = self.xml_soup.find(name=name, attrs=attrs)
            text = self.content.find(name=name, attrs=attrs)
            if text and is_text:
                return text.text
            else:
                return text

    def get_doi(self):
        name = 'idno'
        attrs = {'type': 'DOI'}
        self.doi= self.get_data(name=name,attrs=attrs)

    def get_Abstract(self):
        name = 'abstract'
        self.Abstract= self.get_data(name=name).replace('\n',' ').strip()

    def get_Keywords(self):
        name = 'keywords'
        kwords= self.get_data(name=name)
        if kwords:
            kwords = re.sub('^\n','',kwords)
            kwords = re.sub('\n$','',kwords)
            self.Keywords = kwords.replace('\n',', ').strip()

    def get_Title(self):
        name='title'
        attrs = {'type':'main'}
        self.Title = self.get_data(name=name,attrs=attrs)

    def get_Authors(self):
        name = 'analytic'
        first = self.get_data(name=name, is_text=False)
        name = 'author'
        authors = first.find_all(name=name)
        result = []
        for elt in authors:
            try:
                fname = elt.find(name="forename").text
                lname = elt.find(name="surname").text
                result.append(fname + ' ' + lname)
            except:
                pass
        if result:
            self.Authors = ', '.join(result)
        else:
            self.Authors = ''

    def get_year(self):
        name = 'date'
        attrs = {'type': "published"}
        date = self.get_data(name=name, attrs=attrs)
        if date:
            year = re.findall(regex_paper_year, date)
            if year:
                self.year = year[0]
        else:
            self.year = ''

    def get_Journal(self):
        name='title'
        attrs = {'level': 'j'}
        self.Journal = self.get_data(name=name,attrs=attrs)

    def get_numCitedBy(self):
        if self.doi:
            path = path_open_citations + self.doi
            soup = get_soup(path)
            soup_list = ast.literal_eval(soup.text)
            if isinstance(soup_list,list):
                self.numCitedBy = len(soup_list)

    def get_abbreviations_dict(self) -> dict:
        '''
        Get abbreviations and their meanings from the title and the abstract in a dictionary.
        Modified code from : https://gist.github.com/ijmarshall/b3d1de6ccf4fb8b5ee53
        :return:
        '''
        abstract_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Abstract)
        corrected_abbrevs = {}
        for k in abstract_abbrevs:
            corrected_abbrevs[' ' + k] = ' ' + abstract_abbrevs[k] + ' ' + '(' + k + ')'
        return corrected_abbrevs

    def replace_abbreviations(self):
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)

    def save_dict(self) -> dict:
        '''
        Format the read data of the paper into a dictionary
        :return:
        '''
        attr_to_save = ['doi', 'Authors', 'year', 'database', 'Abstract', 'Title', 'numCitedBy',
                        'Journal']
        paper_to_save = {key: self.__dict__[key] for key in attr_to_save}
        return paper_to_save

class papers_pdf:
    def __init__(self, papers_path):
        '''
        process papers read with grobid
        :param database: name of database used
        :param papers_path:
        :param papers_dict:
        :param nlp:
        '''
        # super().__init__(papers_path, database, nlp)
        self.papers_path = papers_path
        self.elements = {}
        # if papers_path:
        #     self.list_files = [elt for elt in os.listdir(papers_path) if '.xml' in elt]
        # else:
        #     self.list_files = list(papers_dict)
        #     self.papers_dict = papers_dict
        self.list_files = [elt for elt in os.listdir(papers_path) if elt.endswith('.pdf')]
        print('Len data : ', len(self.list_files))

    def read_content(self) -> dict:
        '''
        read pdf files and extract data using grobid.
        :return:
        '''
        pdf_files = [elt for elt in os.listdir(self.papers_path) if elt.endswith('.pdf')]
        data_dict = {}

        for fname in pdf_files:
            path = os.path.join(self.papers_path, fname)
            with open(path, "rb") as f:
                data_dict[fname] = f.read()
        client = GrobidClient(url=grobid_url)
        results = client.process("processHeaderDocument", input_path=data_dict)

        return results

    def add_paper(self,fname,pdf_content):
        new_paper = paper_pdf(fname=fname,pdf_content=pdf_content)
        new_paper.get_doi()
        new_paper.get_Title()
        new_paper.get_Abstract()
        if len(new_paper.Abstract.split()) > 5:
            new_paper.get_Journal()
            new_paper.get_Authors()
            new_paper.get_year()
            new_paper.get_Keywords()
            new_paper.replace_abbreviations()
            new_paper.get_numCitedBy()
            self.elements[fname] = new_paper.save_dict()

    def get_papers(self,show_tqdm=False):
        data = self.read_content()
        if show_tqdm:
            range_ = tqdm(data)
        else:
            range_=data

        for fname in range_:
            self.add_paper(fname=fname,pdf_content=data[fname])




from naimai.constants.nlp import root_dep,verb_pos
from naimai.utils.regex import find_root_verb
from spacy.symbols import VERB, nsubjpass
import re


class active_tense_processor:
    def __init__(self, sentence, nlp):
        self.nlp = nlp
        self.sentence = nlp(sentence)
        self.separators = []
        self.threshold = .7
        self.root_verb = None
        self.root_clause = None
        self.naive_clauses = []

    def get_separators(self):
        result = []
        tokens_to_split = ['which', 'and', 'but', "because", "while", "then", "also", ',', "that"]
        for token in self.sentence:
            if token.text in tokens_to_split:
                result.append(token)
        self.separators = result

    def naive_separator_segmentation(self):
        puncts = [',']
        self.get_separators()
        if self.separators:
            separators_text = [elt.text for elt in list(set(self.separators))]
            separators_uniques = [' ' + elt + ' ' if elt not in puncts else elt for elt in separators_text]
            separators_regex = '|'.join(separators_uniques)
            return re.split(separators_regex, self.sentence.text)
        return [self.sentence.text]

    def is_root_in_clause(self, clause_nlp):
        for token in clause_nlp:
            if token.text == self.root_verb.text:
                return True
        return False

    def idx_clause_with_root(self, first_clauses_list_nlp):
        first_idx = 0
        for clause in first_clauses_list_nlp:
            if self.is_root_in_clause(clause):
                first_idx = first_clauses_list_nlp.index(clause)
                break
        return first_idx

    def get_clauses(self):
        # segment to naive clauses
        naive_clauses = self.naive_separator_segmentation()
        separators = [''] + [token.text for token in self.separators]
        naive_clauses = [self.nlp(sep + ' ' + clause) for sep, clause in zip(separators, naive_clauses)]

        # remove clauses without root verb
        idx_clause_root = self.idx_clause_with_root(naive_clauses)
        self.root_clause = naive_clauses[idx_clause_root]

        # filter clauses with "that + verb"
        for clause in naive_clauses:
            if (clause.text != self.root_clause.text) and (clause[0].text == 'that') and (clause[1].pos == VERB):
                pass
            else:
                self.naive_clauses.append(clause)

    def transform_root_clause(self):
        idx_root_verb = [token.i for token in self.root_clause if token.text == self.root_verb.text][0]
        sub_clause = self.root_clause[idx_root_verb + 1:].text
        # conjugate root verb
        if self.root_verb.tag_ == 'VBZ' or self.root_verb.tag_ == 'VBN':
            root = self.root_verb._.inflect('VBP')
        else:
            root = self.root_verb.text
        new_sentence = 'we ' + root + ' ' + sub_clause
        return self.nlp(new_sentence)

    def transform_active_tense(self):
        first_clause_transformed = self.transform_root_clause()
        # restitution
        return [first_clause_transformed.text] + [clause.text for clause in self.naive_clauses if
                                                  clause.text != self.root_clause.text]

    def process(self):
        # find root verb
        self.root_verb = find_root_verb(sentence=self.sentence)

        # get root & naive clauses
        self.get_clauses()

        result = self.transform_active_tense()

        return ' '.join(result)


class clauses_processor:
  def __init__(self,sentence,nlp):
    self.nlp = nlp
    self.sentence = nlp(sentence)
    self.root_verb=None

  def is_root_passive(self):
    if self.root_verb.tag_=='VBN':
      if 'is' in [child.text for child in self.root_verb.children]:
        return True
    return False

  def find_subject_and_children(self):
    root_verb_subject= [token for token in self.root_verb.children if token.dep==nsubjpass][0]
    subject_children=list(root_verb_subject.subtree)
    subject_children_txt = [token.text for token in subject_children]

    idx_root_verb = [token.i for token in self.sentence if token.text == self.root_verb.text][0]
    rest_after_root_verb = self.sentence[idx_root_verb+1:]
    text_in_beginning = [token for token in list(self.root_verb.subtree) if (token.text not in subject_children_txt) and (token.text not in rest_after_root_verb.text)]

    # remove original root verb:
    for idx,token in enumerate(text_in_beginning):
      if token.text ==self.root_verb.text:
        try:
          del text_in_beginning[idx-1] #delete is
          del text_in_beginning[idx-1] #delete original verb
          break
        except:
          pass
    root_verb_children=subject_children+ [rest_after_root_verb] + text_in_beginning
    root_verb_children =[token.text for token in root_verb_children]
    return root_verb_children

  def passive2active(self):
    rest_of_sentence = self.find_subject_and_children()
    rest_of_sentence = ' '.join(rest_of_sentence)
    root_verb_present_tense = self.root_verb._.inflect('VBP')
    new_sentence = 'we '+ root_verb_present_tense + ' ' + rest_of_sentence
    return new_sentence

  def process_active_tense(self,sentence,nlp):
    processor = active_tense_processor(sentence=sentence, nlp=nlp)
    return processor.process()

  def process(self):
    # find root verb
    self.root_verb = find_root_verb(sentence=self.sentence)

    # if root is active tense format :
    if not self.is_root_passive():
      new_sentence = self.process_active_tense(sentence=self.sentence.text,
                                               nlp = self.nlp)

    else: # if root is passive tense format :
      active_tense = self.passive2active()
      new_sentence = self.process_active_tense(sentence=active_tense,
                                               nlp = self.nlp)
    return new_sentence


class objective_sentence_processor:
    def __init__(self, sentence, nlp):
        self.nlp = nlp
        self.raw_sentence = nlp(sentence)
        self.processed_sentence = None
        self.final_sentence = ''
        self.clauses_processor = None

    # deal with 2 successive verbs & replace beginning with we
    def get_second_verb_when_2_sucessive_verbs(self):
        for token in self.processed_sentence:
            if token.pos_ == verb_pos and token.dep_ == root_dep and token.text != 'used':
                try:
                    next_token = self.processed_sentence[token.i + 1]
                    if next_token.pos_ == 'PART':
                        next_next_token = self.processed_sentence[token.i + 2]
                        if next_next_token.pos_ == verb_pos:
                            return next_next_token
                except:
                    pass
        return

    def fix_2_successive_verbs(self):
        second_verb = self.get_second_verb_when_2_sucessive_verbs()
        if second_verb:
            second_verb_idx = [token.i for token in self.processed_sentence if token == second_verb][0]
            sub_sentence = self.processed_sentence[second_verb_idx + 1:]

            second_verb_text = second_verb.text
            if 'understand' in second_verb_text:
                second_verb_text = 'analyze'

            new_sentence = 'we ' + second_verb_text + ' ' + sub_sentence.text
            self.processed_sentence = self.nlp(new_sentence)

    def get_table_infos(self, processed=True):
        if processed:
            sentence = self.processed_sentence
        else:
            sentence = self.raw_sentence
        print('Token\tIndex\tPOS\tTag\tDep\tAncestor\tChildren')
        for token in sentence:
            ancestors = [t.text for t in token.ancestors]
            children = [t.text for t in token.children]
            print(token.text[:6], "\t", token.i, "\t",
                  token.pos_, "\t",
                  token.tag_, "\t", token.dep_, "\t",
                  ancestors, "\t", children)

    def remove_some_words(self):
        doc_text = re.sub('in particular,?|moreover,?|furthermore,?| also|', '', self.raw_sentence.text, flags=re.I)
        self.processed_sentence = self.nlp(doc_text)

    def replace_synonyms(self):
        pattern = 'together with|along with|as well as|in addition to|besides'
        self.processed_sentence = self.nlp(re.sub(pattern, 'and', self.processed_sentence.text))

    def remove_which_clauses(self):
        # pattern= '[^by which],?\s?(which|through|as\s?a? part of).*?,'
        pattern = '(by which.*)|(,?\s?(which|through|as\s?a? part of).*?,)'
        self.processed_sentence = self.nlp(re.sub(pattern, '', self.processed_sentence.text))

    def remove_numerotations(self):
        doc_text = re.sub(' \((i+v?)\)|\((\d)\)|\(v\)', '', self.processed_sentence.text)
        self.processed_sentence = self.nlp(doc_text)

    def clause_has_verb_or_subject(self, clause):
        root_verb = find_root_verb(sentence=self.processed_sentence)

        subject = [token for token in root_verb.children if token.dep == nsubjpass]
        if subject:
            subject = subject[0].text
        else:
            subject = 'ZZZ'
        for token in clause:
            if (token.pos == VERB) or (token.text == subject):
                return True
        return False

    def remove_first_clause_without_verb(self):
        clauses_by_comma_separator = self.processed_sentence.text.split(',')
        first_clause = self.nlp(clauses_by_comma_separator[0])
        if not self.clause_has_verb_or_subject(first_clause):
            self.processed_sentence = self.nlp(' '.join(clauses_by_comma_separator[1:]))

    def process(self):
        # first processings
        self.remove_some_words()
        # print('some words removed : ', self.processed_sentence)
        self.remove_first_clause_without_verb()
        # print('without first clause removed : ', self.processed_sentence)
        self.remove_which_clauses()
        # print('without which clause removed : ', self.processed_sentence)
        self.remove_numerotations()
        self.replace_synonyms()

        # fix 2 successive verbs
        self.fix_2_successive_verbs()

        # process
        self.clauses_processor = clauses_processor(sentence=self.processed_sentence.text, nlp=self.nlp)
        self.final_sentence = self.clauses_processor.process()


from naimai.papers.full_text.pmc import paper_pmc
import re
from naimai.constants.regex import regex_abstract_omr
from naimai.models.papers_classification.obj_classifier import Objective_classifier
from naimai.constants.regex import regex_filtered_words_obj, regex_objectives, regex_review

class Omr_Review_Paper_Classifier:
    '''
    classify if paper is OMR type or a review
    '''
    def __init__(self,paper: paper_pmc,objective_classifier=None):
        if objective_classifier:
            self.objective_classifier = objective_classifier
        else:
            print('Need GPU to faster things..')
            self.objective_classifier = Objective_classifier()
        self.paper = paper

    def get_paper_abstract(self ) -> str:
        '''
        get abstract paper
        :param paper:
        :return:
        '''
        self.paper.get_Abstract(stacked=False)
        return self.paper.Abstract

    def get_paper_title(self) -> str:
        '''
        get title of paper
        :param paper:
        :return:
        '''
        self.paper.get_Title()
        return self.paper.Title

    def get_paper_text(self) -> str:
        '''
        stack title & abstract in same text
        :param paper:
        :return:
        '''
        title = self.get_paper_title()
        abstract = self.get_paper_abstract()
        text = title + ' ' + abstract
        return text

    def is_text_omr(self ,text) -> bool:
        '''
        detect if the text has omr sections or no based on regex_abstract_omr
        :param text:
        :return:
        '''
        if re.findall(regex_abstract_omr ,text ,re.I):
            return True
        return False

    def find_obj(self,text)->str:
        '''
        identify obj using regex & obj classifier and stack them
        :param text:
        :return:
        '''
        obj_rgx = list(set(re.findall(regex_objectives, text, flags=re.I)))

        list_sentences = text.split('.')
        objs_clf = self.objective_classifier.predict(list_sentences)
        objs_clf = [obj for obj in objs_clf if
                                      not re.findall(regex_filtered_words_obj, obj, flags=re.I)]

        objectives = obj_rgx + objs_clf
        objectives = list(set(objectives))
        if objectives:
            return ' '.join(objectives)
        return ''
    def is_objective_review(self,objective: str) -> bool:
        '''
        objective is considered as review & return True if it contains regex_review. Otherwise it's False (then OMR)
        :param objective:
        :return:
        '''
        if objective:
            if re.findall(regex_review,objective,re.I):
                return True
        return False

    def is_paper_review(self) -> bool:
        '''
        identify is paper is review or not by :
            1. checking if it has omr format, in which case it's not a review paper
            2. check if its obj sentences contains some review terms, in which case it's considered a review paper
        :param paper:
        :return:
        '''
        text = self.get_paper_text()

        if self.is_text_omr(text):
            return False

        objectives = self.find_obj(text)
        is_obj_review = self.is_objective_review(objectives)
        return is_obj_review




from transformers import TrainingArguments, DataCollatorForTokenClassification, EarlyStoppingCallback
from datasets import Dataset
from naimai.utils.general import correct_ner_data
from naimai.utils.transformers import visualize, compute_metrics, get_text
from naimai.constants.models import output_labels
from naimai.constants.paths import path_ner_balanced_data
from .trainer import BOMR_Trainer, Predictions_preparer
from .ner_processor import NER_BOMR_processor
import pandas as pd
import torch


class NER_BOMR_classifier:
    def __init__(self, config={}, path_ner_data=path_ner_balanced_data, ner_data_df=None, model=None, tokenizer=None,label_all_subtokens=False,load_model=False,
                 path_model=None,predict_mode=False, verbose=False):
        self.config = config
        self.tokenized_data = None
        self.trainer = None
        self.datasets = None
        self.label_all_subtokens = label_all_subtokens

        # LABELS ----------------

        self.ids2labels = {k: v for k, v in enumerate(output_labels)}
        self.labels2ids = {v: k for k, v in enumerate(output_labels)}

        # DATA -----------------
        if not predict_mode:
            if isinstance(ner_data_df, pd.DataFrame):
                self.NER_data_df = ner_data_df
            else:
                df = pd.read_csv(path_ner_data)
                df['entities'] = df.apply(correct_ner_data, axis=1)
                df['tokens'] = df['text'].str.split()
                df['ner_tags'] = df['entities'].apply(lambda x: [self.labels2ids[elt] for elt in x])
                self.NER_data_df = df.dropna()
                print(f'NER data shape : {self.NER_data_df.shape}')

        # TOKENIZER ------------------
        if tokenizer:
            self.tokenizer = tokenizer
        elif load_model:
            self.load_model(path_model,verbose=verbose)
        else:
            print('Getting Tokenizer..')
            # self.tokenizer = AutoTokenizer.from_pretrained(config['tokenizer_name'])

        # MODEL  ----------------
        if model:
            self.model = model
        elif load_model:
            pass
        else:
            print('Model creation..')
            # self.model = AutoModelForTokenClassification.from_pretrained(config['model_name'],
            #                                                              num_labels=len(output_labels))

        # Data Collator ------------
        if not predict_mode:
            self.data_collator = DataCollatorForTokenClassification(self.tokenizer)

        # Training ARGS ----------
        if not predict_mode:
            self.training_args = TrainingArguments(
                output_dir=f"{config['output_dir']}",
                save_total_limit = 1,
                evaluation_strategy="steps",
                eval_steps=int(len(self.NER_data_df)*0.8/config['batch_size']),
                metric_for_best_model='eval_F1-avg',
                logging_strategy="epoch",
                save_strategy="steps",
                save_steps=int(len(self.NER_data_df)*0.8/config['batch_size']),
                learning_rate=config['learning_rates'],
                per_device_train_batch_size=config['batch_size'],
                per_device_eval_batch_size=config['batch_size'],
                num_train_epochs=config['epochs'],
                weight_decay=config['weight_decay'],
                load_best_model_at_end=True
            )

    def tokenize_and_align_labels(self, examples):
        tokenized_inputs = self.tokenizer(examples["tokens"], truncation=True, is_split_into_words=True,
                                          padding='max_length',max_length=self.config['max_length'])

        labels = []
        for i, label in enumerate(examples[f"ner_tags"]):
            word_ids = tokenized_inputs.word_ids(batch_index=i)  # Map tokens to their respective word.
            previous_word_idx = None
            label_ids = []
            for word_idx in word_ids:  # Set the special tokens to -100.
                if word_idx is None:
                    label_ids.append(-100)
                elif word_idx != previous_word_idx:  # Only label the first token of a given word.
                    try:
                        label_ids.append(label[word_idx])
                    except:
                        # print(f'idx {i}, widx = {word_idx}, label {label}, len word ids= {len(word_ids)}')
                        label_ids.append(-100)
                else:
                    if self.label_all_subtokens:
                        label_ids.append(label[word_idx])
                    else:
                        label_ids.append(-100)
                previous_word_idx = word_idx
            labels.append(label_ids)


        tokenized_inputs["labels"] = labels
        return tokenized_inputs

    def get_train_validation_data(self, test_size=.20, seed=42):
        ds = Dataset.from_pandas(self.NER_data_df[['doi', 'tokens', 'ner_tags']])
        self.datasets = ds.train_test_split(test_size=test_size, shuffle=True, seed=seed)

    def tokenize_data(self):
        self.tokenized_data = self.datasets.map(self.tokenize_and_align_labels, batched=True)

    def get_trainer(self):
        self.trainer = BOMR_Trainer(
            model=self.model,
            args=self.training_args,
            train_dataset=self.tokenized_data["train"],
            eval_dataset=self.tokenized_data["test"],
            data_collator=self.data_collator,
            tokenizer=self.tokenizer,
            compute_metrics=compute_metrics,
            callbacks=[EarlyStoppingCallback(early_stopping_patience=5)]
        )

    def train(self):
        self.trainer.train()

    def load_model(self,path_model,verbose):
        from transformers import AutoModelForTokenClassification, AutoTokenizer
        if verbose:
            print('  Loading model & tokenizer...')
        self.model = AutoModelForTokenClassification.from_pretrained(path_model, num_labels=len(output_labels))
        self.tokenizer = AutoTokenizer.from_pretrained(path_model)
        if torch.cuda.is_available():
            print('  >> GPU Used in ner classifier !')
            self.model = self.model.to('cuda')
        else:
            print('  >> No GPU used in ner classifier..')

    def predict(self,text,visualize_=True,dict_format=False):
        '''
        Predict sentences class in text and return dict format or in pandas format
        :param text:
        :param visualize_: to visualize results using displacy
        :param dict_format:
        :return:
        '''
        tokens = text.split()
        if torch.cuda.is_available():
            if self.config:
                max_length=self.config['max_length']
            else:
                max_length = 256*5
            encoding = self.tokenizer(tokens, truncation=True, is_split_into_words=True,return_tensors='pt',padding='max_length',max_length=max_length).to('cuda')
        else:
            encoding = self.tokenizer(tokens, truncation=True, is_split_into_words=True, return_tensors='pt')
        outputs = self.model(**encoding)
        prediction = outputs.logits.argmax(-1)[0]

        prp = Predictions_preparer(predictions=prediction,
                                   tokenizer=self.tokenizer,
                                   datasets=None)
        df = prp.prepare_one_prediction("doi", prediction, tokens)

        ner_processor = NER_BOMR_processor(text=text,df=df)
        ner_processor.add_metadata()
        processed_df = ner_processor.df

        visualize(text,processed_df,show=visualize_) #transform df & visualize

        if dict_format:
            processed_df['text']=processed_df.apply(get_text,args=(text,),axis=1)
            prediction_dict = processed_df[['class','text']].set_index('class')['text'].to_dict()
            return prediction_dict
        return processed_df



import sqlite3
import re
from naimai.constants.regex import regex_and_operators,regex_or_operators,regex_exact_match
from ast import literal_eval
from naimai.utils.general import clean_lst

class SQLiteManager:
  def __init__(self,path_db: str):
    connexion = sqlite3.connect(path_db)
    self.cursor = connexion.cursor()

  def get_by_multiple_ids(self,list_ids:  list, year_from=0, year_to=3000,)-> dict:
    '''
      get papers using list of ids
    '''
    tuple_ids = tuple([elt+1 for elt in list_ids])
    self.cursor.execute("SELECT website,year, database, messages, reported, title, journal, authors, numCitedBy, fname, allauthors FROM all_papers WHERE rowid IN {}".format(tuple_ids))
    results = self.cursor.fetchall()
    dict_result = {elt[9]: self.to_dict(elt) for elt in results}
    # filter by years range
    dict_result = {elt: dict_result[elt] for elt in dict_result if dict_result[elt]['year']>year_from and dict_result[elt]['year']<year_to}
    return dict_result

  def search_with_exact_match(self, query: str, year_from=0, year_to=3000,top_n=5) -> dict:
    '''
    find papers for a query with exact match
    '''
    keywords = re.findall(regex_exact_match, query)
    similar_papers = {}

    for word in keywords:
      papers = self.get_by_query(word,year_from=year_from,year_to=year_to,top_n=top_n)
      similar_papers.update(papers)

    return similar_papers

  def search_with_OR_operator(self, query: str, year_from=0, year_to=3000,top_n=5) -> dict:
    '''
    find papers that contains at least one of the query keywords. Similar to many exact match..
    '''
    keywords = [elt.strip() for elt in re.split(regex_or_operators, query)]
    similar_papers = {}

    for word in keywords:
      papers = self.get_by_query(word, year_from=year_from,year_to=year_to,top_n=top_n)
      similar_papers.update(papers)

    return similar_papers

  def search_with_AND_operator(self, query: str, year_from=0, year_to=3000,top_n=5) -> dict:
    '''
    find papers that contains all the query keywords : start by using or operator for first
    key word, then filter papers with all key words
    '''
    keywords = [elt.strip() for elt in re.split(regex_and_operators, query)]
    similar_papers = {}

    kword1 = keywords[0]
    # get papers with only kword1
    papers_with_kword1 = self.get_by_query(kword1,year_from=year_from,year_to=year_to,top_n=top_n)

    # filter the papers having other keywords as well
    pattern = ''.join([f'(?:.*{kw})' for kw in keywords[1:]])
    for fname in papers_with_kword1:
      # get text of paper
      if '_objectives' in fname:
        info = '. '.join(papers_with_kword1[fname]['messages']) + ' ' + papers_with_kword1[fname]['title']
      else:
        info = '. '.join(papers_with_kword1[fname]['messages'])

      # check if contains all keywords
      if re.findall(pattern, info, flags=re.I):
        similar_papers.update({fname: papers_with_kword1[fname]})

    return similar_papers

  def get_by_lemmatized_query(self, lemmatized_query: list, year_from=0, year_to=3000, top_n=5) -> dict:
    '''
    get all papers in range of years using a lemmatized query. The difference with get_by_query is that the lemmatized query is in list format!
    :return:
    '''

    params_lemmatized_query = ['%'+w+'%' for w in lemmatized_query]

    params=[year_to,year_from]+[params_lemmatized_query[i//2] for i in range(len(params_lemmatized_query)*2)]
    kwords_in_msg = ' AND '.join(['(title LIKE ? OR messages LIKE ?)' for _ in lemmatized_query])
    command= "SELECT website,year, database, messages, reported, title, journal, authors, numCitedBy, fname, allauthors FROM all_papers WHERE (year < ? AND year > ?) AND "+kwords_in_msg

    self.cursor.execute(command, params)
    result = self.cursor.fetchmany(top_n)
    papers_list = clean_lst(result)
    dict_result = {elt[9]: self.to_dict(elt) for elt in papers_list}
    if len(lemmatized_query)==1:
      query = lemmatized_query[0]
      filtered_dict_result = self.filter_papers(query,dict_result)
      return filtered_dict_result
    return dict_result

  def filter_papers(self,query: str,papers: dict) -> dict:
    '''
    filter papers when the query is in the word
    :param query:
    :param papers:
    :return:
    '''
    pattern = '[^a-zA-Z]' + query
    new_papers = {}
    for fname in papers:
      messages = papers[fname]['messages']
      for text in messages:
        if re.findall(pattern,text,re.I):
          new_papers[fname]=papers[fname]
          break
    return new_papers

  def get_by_query(self, query: str, year_from=0, year_to=3000,top_n=5) -> dict:
    '''
    get papers that contains query in title or message. Here, the input query is not lemmatized.
    :param query:
    :return:
    '''
    query_command = '%' + query + '%'
    qry = (year_to,year_from,query_command,query_command)
    self.cursor.execute("SELECT website,year, database, messages, reported, title, journal, authors, numCitedBy, fname, allauthors FROM all_papers WHERE (year < ? AND year > ?) AND (title LIKE ? OR messages LIKE ?) ",qry)
    # result = self.cursor.fetchall()
    result = self.cursor.fetchmany(top_n)
    papers_list = clean_lst(result)
    dict_result = {elt[9]: self.to_dict(elt) for elt in papers_list}
    filtered_dict_result = self.filter_papers(query,dict_result)

    return filtered_dict_result

  def get_by_fname(self, fname: str, year_from=0, year_to=3000) -> dict:
    '''
      find paper by fname
    '''
    self.cursor.execute(
      "SELECT website,year, database, messages, reported, title, journal, authors, numCitedBy, fname, allauthors FROM all_papers WHERE fname = ?",
      (fname,))
    result = self.cursor.fetchone()
    dict_result = self.to_dict(result)
    if fname.endswith('_objectives'):
      if 'year' in dict_result:
        if (int(dict_result['year']) >= year_from) and (int(dict_result['year']) <= year_to):
          return dict_result
      return {}
    return dict_result

  def get_omr_dics(self,fname: str) -> dict:
    '''
    for fname, get obj, methods & results dict
    '''
    paper_name = '_'.join(fname.split('_')[:-1])
    omr_fnames = [paper_name+'_objectives', paper_name+'_methods',paper_name+'_results']

    result = {}
    for fname in omr_fnames:
      paper = self.get_by_fname(fname)
      result[fname]= paper
    return result


  # def get_by_multiple_fnames(self,fnames: list,year_from=0,year_to=3000) -> dict:
  #   '''
  #     get papers between range of years using list of filenames
  #   '''
  #   params = [year_to,year_from] + clean_lst(fnames)
  #   ln = len(fnames)
  #   conditions = '('+ ' OR '.join(["fname = ?"] * ln)  + ')'
  #   command = "SELECT website,year, database, messages, reported, title, journal, authors, numCitedBy, fname, allauthors FROM all_papers WHERE (year < ? AND year > ?) AND " + conditions
  #   self.cursor.execute(command, params)
  #   list_papers = self.cursor.fetchall()
  #   dict_result = {elt[9]: self.to_dict(elt) for elt in list_papers}
  #   return dict_result


  def to_dict(self,sql_result: tuple) -> dict:
    '''
      transform sql results (in tuple format) to dictionary format
    '''
    cols = ['website','year', 'database', 'messages', 'reported', 'title', 'journal',
    'authors', 'numCitedBy', 'fname','allauthors']

    dict_result = {}
    if sql_result:
      for col,val in zip(cols,sql_result):
        if val:
          try:
            if val[0]=='[' and val[-1]==']':
              val = literal_eval(val)
            dict_result[col]= val
          except:
            pass
      return dict_result
    return {}



nlp_disabled_processing = ["tok2vec", "parser", "attribute_ruler", 'entity_ruler','textcat','textcar_multilabel','morphologizer','senticizer','transformer']
nlp_disabled_direct2reported = ["tok2vec", "attribute_ruler", 'entity_ruler','textcat','textcar_multilabel','morphologizer','senticizer','transformer', "ner"]

nlp_vocab = 'en_core_web_sm'

past_tense='VBD'
verb_pos='VERB'
root_dep='ROOT'
conj_dep = 'conj'

this_year=2021
max_len_objective_sentence = 45


from naimai.constants.regex import regex_filtered_words_obj
from naimai.constants.nlp import nlp_vocab
from naimai.constants.paths import path_bomr_classifier
from naimai.utils.regex import get_first_last_token_ids
from naimai.models.papers_classification.bomr_classif.ner_processor import NER_BOMR_processor
from naimai.models.papers_classification.bomr_classif.ner_classifier import NER_BOMR_classifier
from naimai.models.papers_classification.obj_classifier import Objective_classifier
import re
import spacy


def segment(text, obj_clf=None, bomr_clf=None, path_bomr_clf=path_bomr_classifier, summarize=True, check_bg=False, visualize_=True,nlp=None):
    '''
    segment a text using OMR_Text_Segmentor.
    :param text:
    :param obj_clf:
    :param bomr_clf:
    :param summarize:
    :param check_bg:
    :param visualize_:
    :param filter_bg:
    :return:
    '''
    segmentor = OMR_Text_Segmentor(text=text, objective_classifier=obj_clf,
                                   bomr_classifier=bomr_clf, summarize=summarize,path_bomr_clf=path_bomr_clf)
    if visualize_:
        segmentor.bomr_classifier.predict(text=segmentor.text, visualize_=True, dict_format=False)

    segmented = segmentor.segment(check_bg=check_bg)
    background_in = 'background' in segmented.values()
    if background_in:
        stcs = [segmentor.sentences[idx] for idx in segmented if segmented[idx] != 'background']
        summarized = ' '.join(stcs)
        segmentor = OMR_Text_Segmentor(text=summarized, objective_classifier=segmentor.objective_classifier,
                                      bomr_classifier=segmentor.bomr_classifier, summarize=False,nlp=nlp)
        if visualize_:
            print('')
            print('visualization of summarized : ')
            segmentor.bomr_classifier.predict(text=segmentor.text, visualize_=True, dict_format=False)

        segmented = segmentor.segment(check_bg=False)
    return (segmented,segmentor.sentences,segmentor.objective_classifier,segmentor.bomr_classifier)

class OMR_Text_Segmentor:
    '''
    Segment a text to Objective, Method and Results
    '''
    def __init__(self, text, bomr_classifier=None, objective_classifier=None, path_bomr_clf=path_bomr_classifier,nlp=None,summarize=False):
        self.text = text
        self.ner_processor = None
        # self.range_wids_labels=None

        # Segment to sentences
        if not nlp:
            nlp = spacy.load(nlp_vocab)
        doc = nlp(text)
        regex_online_version = 'online version'
        self.sentences = [sent.text.strip() for sent in doc.sents if len(sent.text.split())>5 and not re.findall(regex_online_version,sent.text,flags=re.I)]
        self.text = ' '.join(self.sentences)

        # Get obj classifier
        self.objective_classifier = None
        self.load_obj_classifier(objective_classifier)

        if summarize:
            self.summarize()

        # Get bomr classifier
        self.bomr_classifier=None
        self.load_bomr_classifier(bomr_classifier,path=path_bomr_clf)

    def load_obj_classifier(self, obj_classifier):
        if obj_classifier:
            self.objective_classifier = obj_classifier
        else:
            print('>> Loading obj classifier..')
            self.objective_classifier = Objective_classifier()

    def load_bomr_classifier(self, bomr_classifier,path=path_bomr_classifier):
        if bomr_classifier:
            self.bomr_classifier = bomr_classifier
        else:
            print('>> Loading bomr classifier..')
            self.bomr_classifier = NER_BOMR_classifier(load_model=True, path_model=path,
                                                       predict_mode=True, verbose=False)

    def text2bomr(self, text, visualize_=False):
        return self.bomr_classifier.predict(text=text, visualize_=visualize_, dict_format=False)

    def get_ner_processor(self,text):
        df = self.text2bomr(text)
        return NER_BOMR_processor(text=self.text, df=df)

    def get_sentence_label(self, sentence, ner_processor,range_wids_labels,last_token=0):
        sentence_range_wids = get_first_last_token_ids(sentence, self.text, last_token)
        # if not self.range_wids_labels:

        sentence_class = ner_processor.get_label_with_most_overlap(sentence_range_wids=sentence_range_wids,
                                                                        range_wids_labels=range_wids_labels)
        return (sentence_class, sentence_range_wids[-1] + 1)

    def segment_naive(self) -> dict:
        '''
        segment text into bomr using longformer/bigbird predictions & ner processor
        :return:
        '''
        last_tokenId = 0
        segmented_text = {}
        ner_processor = self.get_ner_processor(self.text)
        range_wids_labels = ner_processor.get_range_wids_labels()
        for idx, sentence in enumerate(self.sentences):
            segmented_text[idx], last_tokenId = self.get_sentence_label(sentence=sentence,
                                                                        ner_processor=ner_processor,
                                                                        range_wids_labels=range_wids_labels,
                                                                        last_token=last_tokenId)
        return segmented_text

    def summarize(self):
        '''
        filter sentences using objective classifier
        :return:
        '''
        sentences= self.get_objective_sentences(self.sentences)
        if len(sentences)>3:
            self.sentences=sentences
            self.text = ' '.join(self.sentences)

    def get_objective_sentences(self,sentences: list):
        objectives = self.objective_classifier.predict(sentences)
        objectives_with_classifier = [obj for obj in objectives if
                                      not re.findall(regex_filtered_words_obj, obj, flags=re.I)]
        return objectives_with_classifier

    def filter_background(self,segmented_text: dict)-> dict:
        '''
         process each background classes : get matched objective sentences with objective classifier.
        :param segmented_text:
        :return:
        '''
        regex_bground = 'background'
        bground_sentences = [self.sentences[idx] for idx in segmented_text if
                             re.findall(regex_bground, segmented_text[idx])]

        objectives_with_classifier = self.get_objective_sentences(bground_sentences)
        obj_sentences_ids = [self.sentences.index(elt) for elt in objectives_with_classifier]

        corrected = {}
        for idx in obj_sentences_ids:
            corrected[idx] = 'objectives'

        for idx in segmented_text:
            if not re.findall(regex_bground, segmented_text[idx]):
                corrected[idx] = segmented_text[idx]
        return corrected

    def segment(self,check_bg=False) -> dict:
        '''
        segment naively the text & if check_bg, extract objectives from bground using filter_background.
        if remove_bg_resegment, remove background from segmentation and recompute.
        :param check_background:
        :return:
        '''
        segmented_text = self.segment_naive()
        if check_bg:
            segmented_text = self.filter_background(segmented_text)
        #
        # if remove_bg_resegment:
        #     segmented_text = self.segment_without_bground(segmented_text)

        return segmented_text




'''
Modified grobid client from Grobid source: https://github.com/kermitt2/grobid. The GrobidClient object
takes a directory of PDFs and converts it to XML files as explained in the process method.
'''
import os
import json
import time
import concurrent.futures
import requests
from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import shutil

from .client import ApiClient


class ServerUnavailableException(Exception):
    pass


class GrobidClient(ApiClient):

    def __init__(self, url,
                 coordinates=["persName", "figure", "ref", "biblStruct", "formula", "s"],
                 sleep_time=5,
                 timeout=180,
                 config_path=None,
                 check_server=True):
        self.config = {
            'url': url,
            'coordinates': coordinates,
            'sleep_time': sleep_time,
            'timeout': timeout
        }
        if config_path:
            self._load_config(config_path)
        if check_server:
            self._test_server_connection()

    def _load_config(self, path="./config.json"):
        """
        Load the json configuration
        """
        config_json = open(path).read()
        self.config = json.loads(config_json)

    def _test_server_connection(self):
        """Test if the server is up and running."""
        the_url = self.config['url']
        r = requests.get(the_url)

        status = r.status_code

        if status != 200:
            print("GROBID server does not appear up and running " + str(status))
        else:
            print("GROBID server is up and running")

    def process(
            self,
            service,
            input_path: str,
            n=10,
            generateIDs=False,
            consolidate_header=True,
            consolidate_citations=False,
            include_raw_citations=False,
            include_raw_affiliations=False,
            tei_coordinates=False,
            segment_sentences=False,
            verbose=False,
            path_export='',
            path_export_new_fnames_pdfs='',
            idx_start=0,
            idx_finish=-1
    ):
        '''
        Process a path of PDFs (input_path) and converts the files into XML files exported in
        path_export.

        :param input_path: of PDF files
        :param path_export: where the PDF files are exported in XML format
        :param path_export_new_fnames_pdfs: where the XML files are exported with title as filename
        :param idx_start: can process all the files in input_path starting from idx_start^th file
        :param idx_finish: can process all the files in input_path until the idx_start^th file
        :return:
        '''

        input_files = []
        result = {}
        if idx_finish==-1:
            filenames = sorted(os.listdir(input_path))[idx_start:]
        else:
            filenames = sorted(os.listdir(input_path))[idx_start:idx_finish]

        for filename in tqdm(filenames):
            if filename.endswith(".pdf") or filename.endswith(".PDF") or \
                    (service == 'processCitationList' and (filename.endswith(".txt") or filename.endswith(".TXT"))):
                if verbose:
                    try:
                        print(filename)
                    except Exception:

                        pass
                input_files.append(os.sep.join([input_path, filename]))

                result[filename] = self.process_batch(
                    service,
                    input_files,
                    n,
                    generateIDs,
                    consolidate_header,
                    consolidate_citations,
                    include_raw_citations,
                    include_raw_affiliations,
                    tei_coordinates,
                    segment_sentences,
                    verbose,
                )
                input_files = []

        if path_export:
            print('Files exported in :', path_export)
            for fname in result:
                text = result[fname]
                title = self.get_paper_title(text)
                if title:
                    title = title.replace('/','_')
                    fle_name = title+'.xml'
                    new_pdf_name = title+'.pdf'
                    output_pdf_path = os.path.join(path_export_new_fnames_pdfs, new_pdf_name)
                    intput_pdf_dir = os.path.join(input_path,fname)
                    shutil.copyfile(intput_pdf_dir, output_pdf_path)
                else:
                    fle_name = fname.replace('pdf','xml')
                output_fname = os.path.join(path_export,fle_name)


                with open(output_fname,'w',encoding='utf8') as xml_file:
                    xml_file.write(text)

        return result

    def get_paper_title(self,text):
        name = 'title'
        attrs = {'type': 'main'}

        soup = BeautifulSoup(text, "lxml")
        content = soup.find(name=name, attrs=attrs)
        if content:
            return content.text
        return
    def process_batch(
            self,
            service,
            input_files,
            n,
            generateIDs,
            consolidate_header,
            consolidate_citations,
            include_raw_citations,
            include_raw_affiliations,
            tei_coordinates,
            segment_sentences,
            verbose=False,
    ):
        if verbose:
            print(len(input_files), "files to process in current batch")

        # we use ThreadPoolExecutor and not ProcessPoolExecutor because it is an I/O intensive process
        with concurrent.futures.ThreadPoolExecutor(max_workers=n) as executor:
            # with concurrent.futures.ProcessPoolExecutor(max_workers=n) as executor:
            results = []
            for input_file in input_files:
                selected_process = self.process_pdf
                if service == 'processCitationList':
                    selected_process = self.process_txt

                r = executor.submit(
                    selected_process,
                    service,
                    input_file,
                    generateIDs,
                    consolidate_header,
                    consolidate_citations,
                    include_raw_citations,
                    include_raw_affiliations,
                    tei_coordinates,
                    segment_sentences)

                results.append(r)
        for r in concurrent.futures.as_completed(results):
            input_file, status, text = r.result()
            # filename = self._output_file_name(input_file, input_path, output)

            if text is None:
                print("Processing of", input_file, "failed with error", str(status))
                return ''
            else:
                return text

    def process_pdf(
            self,
            service,
            pdf_file,
            generateIDs,
            consolidate_header,
            consolidate_citations,
            include_raw_citations,
            include_raw_affiliations,
            tei_coordinates,
            segment_sentences
    ):
        if isinstance(pdf_file,bytes):
            files = {
                "input": (
                    pdf_file,
                    pdf_file,
                    "application/pdf",
                    {"Expires": "0"},
                )
            }
        else:
            files = {
                "input": (
                    pdf_file,
                    open(pdf_file, "rb"),
                    "application/pdf",
                    {"Expires": "0"},
                )
            }
        the_url = self.config['url']
        the_url += "/api/" + service

        # set the GROBID parameters
        the_data = {}
        if generateIDs:
            the_data["generateIDs"] = "1"
        if consolidate_header:
            the_data["consolidateHeader"] = "1"
        if consolidate_citations:
            the_data["consolidateCitations"] = "1"
        if include_raw_citations:
            the_data["includeRawCitations"] = "1"
        if include_raw_affiliations:
            the_data["includeRawAffiliations"] = "1"
        if tei_coordinates:
            the_data["teiCoordinates"] = self.config["coordinates"]
        if segment_sentences:
            the_data["segmentSentences"] = "1"

        try:

            res, status = self.post(
                url=the_url, files=files, data=the_data, headers={"Accept": "text/plain"},
                timeout=self.config['timeout']
            )

            if status == 503:
                time.sleep(self.config["sleep_time"])
                return self.process_pdf(
                    service,
                    pdf_file,
                    generateIDs,
                    consolidate_header,
                    consolidate_citations,
                    include_raw_citations,
                    include_raw_affiliations,
                    tei_coordinates,
                    segment_sentences
                )
        except requests.exceptions.ReadTimeout:
            return (pdf_file, 408, None)

        return (pdf_file, status, res.text)

    def process_txt(
            self,
            service,
            txt_file,
            generateIDs,
            consolidate_header,
            consolidate_citations,
            include_raw_citations,
            include_raw_affiliations,
            tei_coordinates,
            segment_sentences
    ):
        # create request based on file content
        references = None
        with open(txt_file) as f:
            references = [line.rstrip() for line in f]

        the_url = self.config['url']
        the_url += "/api/" + service

        # set the GROBID parameters
        the_data = {}
        if consolidate_citations:
            the_data["consolidateCitations"] = "1"
        if include_raw_citations:
            the_data["includeRawCitations"] = "1"
        the_data["citations"] = references
        res, status = self.post(
            url=the_url, data=the_data, headers={"Accept": "application/xml"}
        )

        if status == 503:
            time.sleep(self.config["sleep_time"])
            return self.process_txt(
                service,
                txt_file,
                generateIDs,
                consolidate_header,
                consolidate_citations,
                include_raw_citations,
                include_raw_affiliations,
                tei_coordinates,
                segment_sentences
            )

        return (txt_file, status, res.text)




fields_codes_elsevier={'Chemistry':['1500-1508','1600-1606'],
           'Artificial Intelligence': ['1702','1707','1711'],
           'Life and Health Science': ['1100-1111','1502','1300-1315', '1911','2204','2307','2400-2406','2700-2748', '2800-2809','2900-2923','3000-3005','3400-3404','3500-3506','3600-3616'],
           'Environmental Science and Renewable Energy': ['1900-1910','1913','2103','2105','2300-2306','2312','2309-2312'],
           'Material Science': ['2211','2500-2508'],
           'Physics': ['1506','1507','3100-3110','1607','2100-2102','2104'],
           'Mathematics': ['1703','2600-2614'],
           'Social Sciences': ['1200-1213','1400-1410','1800-1804','2000-2003','2308','3200-3207','3300-3322'],
           'Computer Science': ['1700-1712'],
           'Engineering': ['1500-1502', '1909', '2102','2104','2200-2216', '1711','1603'],
           'Economics and Finance': ['1804','2000-2003'],
           'Multidisciplinary':['1000']}

all_fields=['Economics and Finance',
 'Chemistry',
 'Artificial Intelligence',
 'Life and Health Science',
 'Material Science',
 'Physics',
 'Mathematics',
 'Social Sciences',
 'Computer Science',
 'Engineering',
 'Environmental Science and Renewable Energy',]


arxiv_fields_abbrevs = {'cs': 'Computer Science',
 'econ': 'Economics',
 'eess': 'Electrical Engineering and Systems Science',
 'math': 'Mathematics',
 'astro-ph': 'Astrophysics',
 'cond-mat': 'Condensed Matter',
 'gr-qc': 'General Relativity and Quantum Cosmology',
 'hep-ex': 'High Energy Physics - Experiment',
 'hep-lat': 'High Energy Physics - Lattice',
 'hep-ph': 'High Energy Physics - Phenomenology',
 'hep-th': 'High Energy Physics - Theory',
 'math-ph': 'Mathematical Physic',
 'nlin' : 'Nonlinear Sciences',
 'nucl-ex': 'Nuclear Experiment',
 'nucl-th': 'Nuclear Theory',
 'physics': 'physics',
 'quant-ph': 'Quantum Physics',
 'q-bio': 'Quantitative Biology',
 'q-fin': 'Quantitative Finance',
 'stat': 'Statistics'}

arxiv_cs_categories =  {'cs.AI' : 'Artificial Intelligence',
 'cs.AR' : 'Hardware Architecture',
 'cs.CC' : 'Computational Complexity',
 'cs.CE' : 'Computational Engineering, Finance, and Science',
 'cs.CG' : 'Computational Geometry',
 'cs.CL' : 'Computation and Language',
 'cs.CR' : 'Cryptography and Security',
 'cs.CV' : 'Computer Vision and Pattern Recognition',
 'cs.CY' : 'Computers and Society',
 'cs.DB' : 'Databases',
 'cs.DC' : 'Distributed, Parallel, and Cluster Computing',
 'cs.DL' : 'Digital Libraries',
 'cs.DM' : 'Discrete Mathematics',
 'cs.DS' : 'Data Structures and Algorithms',
 'cs.ET' : 'Emerging Technologies',
 'cs.FL' : 'Formal Languages and Automata Theory',
 'cs.GL' : 'General Literature',
 'cs.GR' : 'Graphics',
 'cs.GT' : 'Computer Science and Game Theory',
 'cs.HC' : 'Human-Computer Interaction',
 'cs.IR' : 'Information Retrieval',
 'cs.IT' : 'Information Theory',
 'cs.LG' : 'Machine Learning',
 'cs.LO' : 'Logic in Computer Science',
 'cs.MA' : 'Multiagent Systems',
 'cs.MM' : 'Multimedia',
 'cs.MS' : 'Mathematical Software',
 'cs.NA' : 'Numerical Analysis',
 'cs.NE' : 'Neural and Evolutionary Computing',
 'cs.NI' : 'Networking and Internet Architecture',
 'cs.OH' : 'Other Computer Science',
 'cs.OS' : 'Operating Systems',
 'cs.PF' : 'Performance',
 'cs.PL' : 'Programming Languages',
 'cs.RO' : 'Robotics',
 'cs.SC' : 'Symbolic Computation',
 'cs.SD' : 'Sound',
 'cs.SE' : 'Software Engineering',
 'cs.SI' : 'Social and Information Networks',
 'cs.SY' : 'Systems and Control'}
arxiv_econ_categories={'econ.EM': 'Econometrics',
 'econ.GN': 'General Economics',
 'econ.TH': 'Theoretical Economics'}
arxiv_eess_categories = {'eess.AS': 'Audio and Speech Processing',
'eess.IV': 'Image and Video Processing',
'eess.SP': 'Signal Processing',
'eess.SY': 'Systems and Control'}
arxiv_math_categories = {'math.AC': 'Commutative Algebra',
 'math.AG': 'Algebraic Geometry',
 'math.AP': 'Analysis of PDEs',
 'math.AT': 'Algebraic Topology',
 'math.CA': 'Classical Analysis and ODEs',
 'math.CO': 'Combinatorics',
 'math.CT': 'Category Theory',
 'math.CV': 'Complex Variables',
 'math.DG': 'Differential Geometry',
 'math.DS': 'Dynamical Systems',
 'math.FA': 'Functional Analysis',
 'math.GM': 'General Mathematics',
 'math.GN': 'General Topology',
 'math.GR': 'Group Theory',
 'math.GT': 'Geometric Topology',
 'math.HO': 'History and Overview',
 'math.IT': 'Information Theory',
 'math.KT': 'K-Theory and Homology',
 'math.LO': 'Logic',
 'math.MG': 'Metric Geometry',
 'math.MP': 'Mathematical Physics',
 'math.NA': 'Numerical Analysis',
 'math.NT': 'Number Theory',
 'math.OA': 'Operator Algebras',
 'math.OC': 'Optimization and Control',
 'math.PR': 'Probability',
 'math.QA': 'Quantum Algebra',
 'math.RA': 'Rings and Algebras',
 'math.RT': 'Representation Theory',
 'math.SG': 'Symplectic Geometry',
 'math.SP': 'Spectral Theory',
 'math.ST': 'Statistics Theory'}
arxiv_astro_categories = { 'astro-ph.CO': 'Cosmology and Nongalactic Astrophysics',
 'astro-ph.EP': 'Earth and Planetary Astrophysics',
 'astro-ph.GA': 'Astrophysics of Galaxies',
 'astro-ph.HE': 'High Energy Astrophysical Phenomena',
 'astro-ph.IM': 'Instrumentation and Methods for Astrophysics',
 'astro-ph.SR': 'Solar and Stellar Astrophysics'}
arxiv_cond_categories = {'cond-mat.dis-nn' : 'Disordered Systems and Neural Networks',
'cond-mat.mes-hall' : 'Mesoscale and Nanoscale Physics',
'cond-mat.mtrl-sci' : 'Materials Science',
'cond-mat.other' : 'Other Condensed Matter',
'cond-mat.quant-gas' : 'Quantum Gases',
'cond-mat.soft' : 'Soft Condensed Matter',
'cond-mat.stat-mech' : 'Statistical Mechanics',
'cond-mat.str-el' : 'Strongly Correlated Electrons',
'cond-mat.supr-con' : 'Superconductivity'}
arxiv_grqc_categories = {'gr-qc': 'General Relativity and Quantum Cosmology'}
arxiv_hepex_categories = {'hep-ex': 'High Energy Physics - Experiment'}
arxiv_heplat_categories = {'hep-lat' : 'High Energy Physics - Lattice'}
arxiv_hepph_categories = {'hep-ph': 'High Energy Physics - Phenomenology'}
arxiv_hepth_categories = {'hep-th' : 'High Energy Physics - Theory'}
arxiv_mathph_categories = {'math-ph' : 'Mathematical Physics'}
arxiv_nonlinear_categories = {'nlin.AO' : 'Adaptation and Self-Organizing Systems',
 'nlin.CD' : 'Chaotic Dynamics',
 'nlin.CG' : 'Cellular Automata and Lattice Gases',
 'nlin.PS' : 'Pattern Formation and Solitons',
 'nlin.SI' : 'Exactly Solvable and Integrable Systems'}
arxiv_nuclex_categories = {'nucl-ex': 'Nuclear Experiment'}
arxiv_nuclth_categories = {'nucl-th': 'Nuclear Theory'}
arxiv_physics_categories = {'physics.acc-ph' : 'Accelerator Physics',
'physics.ao-ph' : 'Atmospheric and Oceanic Physics',
'physics.app-ph' : 'Applied Physics',
'physics.atm-clus' : 'Atomic and Molecular Clusters',
'physics.atom-ph' : 'Atomic Physics',
'physics.bio-ph' : 'Biological Physics',
'physics.chem-ph' : 'Chemical Physics',
'physics.class-ph' : 'Classical Physics',
'physics.comp-ph' : 'Computational Physics',
'physics.data-an' : 'Data Analysis, Statistics and Probability',
'physics.ed-ph' : 'Physics Education',
'physics.flu-dyn' : 'Fluid Dynamics',
'physics.gen-ph' : 'General Physics',
'physics.geo-ph' : 'Geophysics',
'physics.hist-ph' : 'History and Philosophy of Physics',
'physics.ins-det' : 'Instrumentation and Detectors',
'physics.med-ph' : 'Medical Physics',
'physics.optics' : 'Optics',
'physics.plasm-ph' : 'Plasma Physics',
'physics.pop-ph' : 'Popular Physics',
'physics.soc-ph' : 'Physics and Society',
'physics.space-ph' : 'Space Physics'}
arxiv_quantph_categories = {'quant-ph': 'Quantum Physics'}
arxiv_qbio_categories = {'q-bio.BM' : 'Biomolecules',
'q-bio.CB' : 'Cell Behavior',
'q-bio.GN' : 'Genomics',
'q-bio.MN' : 'Molecular Networks',
'q-bio.NC' : 'Neurons and Cognition',
'q-bio.OT' : 'Other Quantitative Biology',
'q-bio.PE' : 'Populations and Evolution',
'q-bio.QM' : 'Quantitative Methods',
'q-bio.SC' : 'Subcellular Processes',
'q-bio.TO' : 'Tissues and Organs'}
arxiv_qfin_categories = {'q-fin.CP': 'Computational Finance',
'q-fin.EC': 'Economics',
'q-fin.GN': 'General Finance',
'q-fin.MF': 'Mathematical Finance',
'q-fin.PM': 'Portfolio Management',
'q-fin.PR': 'Pricing of Securities',
'q-fin.RM': 'Risk Management',
'q-fin.ST': 'Statistical Finance',
'q-fin.TR': 'Trading and Market Microstructure'}
arxiv_stat_categories = {'stat.AP': 'Applications',
'stat.CO': 'Computation',
'stat.ME': 'Methodology',
'stat.ML': 'Machine Learning',
'stat.OT': 'Other Statistics',
'stat.TH': 'Statistics Theory'}

arxiv_fields_categories = {'cs': arxiv_cs_categories,
                           'econ': arxiv_econ_categories,
                           'eess': arxiv_eess_categories,
                           'math': arxiv_math_categories,
                           'astro-ph': arxiv_astro_categories,
                           'cond-mat': arxiv_cond_categories,
                           'gr-qc': arxiv_grqc_categories,
                           'hep-ex': arxiv_hepex_categories,
                           'hep-lat': arxiv_heplat_categories,
                           'hep-ph': arxiv_hepph_categories,
                           'hep-th': arxiv_hepth_categories,
                           'math-ph': arxiv_mathph_categories,
                           'nlin': arxiv_nonlinear_categories,
                           'nucl-ex': arxiv_nuclex_categories,
                           'nucl-th': arxiv_nuclth_categories,
                           'physics': arxiv_physics_categories,
                           'quant-ph': arxiv_quantph_categories,
                           'q-bio': arxiv_qbio_categories,
                           'q-fin': arxiv_qfin_categories,
                           'stat': arxiv_stat_categories}



from spacy.symbols import VERB, dobj, nsubjpass, pobj, auxpass, xcomp, advcl, prep, ccomp
import re
import pyinflect
import spacy
from naimai.constants.nlp import past_tense, nlp_vocab,root_dep,conj_dep, nlp_disabled_direct2reported
from naimai.constants.regex import study_terms, regex_arxiv_filename
#from naimai.constants.paths import aws_root_pdfs
from naimai.constants.replacements import reported_corrections
from naimai.utils.regex import multiple_replace, reinsert_commas


class Direct2Reported:
    def __init__(self, authors, sentence, nlp=None):
        if nlp:
            self.nlp = nlp
            self.sentence = self.nlp(sentence)
        else:
            self.nlp = spacy.load(nlp_vocab)
            self.sentence = self.nlp(sentence, disable=nlp_disabled_direct2reported)
        self.dependencies = [nsubjpass, dobj, pobj, advcl, prep, ccomp]
        self.verb_dependencies = [root_dep, conj_dep]
        self.corrections_replacements = reported_corrections
        self.authors = authors
        self.action_verbs = []
        self.bows = {}
        self.reported = ''

    def tokenize(self):
        sentence_split = re.findall(r"[\w']+|[.,!?;]", self.sentence.text)
        return sentence_split

    def select_verb_in_2vbs(self, id_verb_to_remove, sentence_split):
        to_remove = id_verb_to_remove[0]
        if 'and' == sentence_split[to_remove - 1]:
            del sentence_split[to_remove - 1:to_remove + 1]
        return self.nlp(' '.join(sentence_split))

    def select_verb_in_3vbs(self, id_verb_to_remove, sentence_split):
        to_remove = id_verb_to_remove[0]
        if 'and' == sentence_split[to_remove - 1]:
            del sentence_split[to_remove - 4:to_remove + 1]
        return self.nlp(' '.join(sentence_split))

    def select_verbs(self):
        all_verbs = self.get_all_verbs(in_str=True)
        sentence_split = self.tokenize()
        verbs_idxs = [sentence_split.index(elt) for elt in all_verbs]

        try:
            if len(verbs_idxs) > 0:
                id_verb_to_remove_in_2vbs = [verbs_idxs[i + 1] for i in range(len(verbs_idxs) - 1) if
                                             verbs_idxs[i + 1] == verbs_idxs[
                                                 i] + 2]  # action_verb1, action verb2, and action_verb3 -> action_verb1
                id_verb_to_remove_in_3vbs = [verbs_idxs[i + 1] for i in range(len(verbs_idxs) - 1) if
                                             (verbs_idxs[i + 1] == verbs_idxs[i] + 3) and (verbs_idxs[i] == verbs_idxs[
                                                 i - 1] + 2)]  # action_verb1 and action verb1 -> action_verb1

                if id_verb_to_remove_in_2vbs:
                    self.sentence = self.select_verb_in_2vbs(id_verb_to_remove_in_2vbs, sentence_split)

                if id_verb_to_remove_in_3vbs:
                    self.sentence = self.select_verb_in_3vbs(id_verb_to_remove_in_3vbs, sentence_split)

        except:
            pass

    def get_all_verbs(self, in_str=False):
        result = []
        if in_str:
            result = [token.text for token in self.sentence if token.pos == VERB]
        else:
            result = [token for token in self.sentence if token.pos == VERB]
        return result

    def get_bow_of_verb(self, verb):
        right_words = self.get_words_verb(verb)
        right_words_children = []
        # get the bow in the right dependency direction
        if right_words:
            for wd in right_words:
                right_words_children += [child for child in wd.subtree]
            return right_words_children
        # else:
        #     print('No next word after verb in ', self.sentence)

    def correct_bow(self):
        # in case words of second verb were considered in the first verb two, they're removed
        first_verb = self.action_verbs[0]
        if len(self.bows) == 2:
            second_verb = self.action_verbs[1]
            self.bows[first_verb] = [elt for elt in self.bows[first_verb] if elt not in self.bows[second_verb]]
            try:
                self.bows[first_verb].remove(second_verb)
            except:
                pass

    def get_bow(self):
        self.select_verbs()
        self.get_action_verbs()

        if self.action_verbs:
            # get right direction using dependencies
            for verb in self.action_verbs:
                self.bows[verb] = self.get_bow_of_verb(verb)

            self.correct_bow()
        # else:
        #     print('No action verbs in ', self.sentence)

    def sort_bows(self):
        for verb in self.bows:
            positions = sorted(self.bows[verb], key=lambda word: word.idx)
            self.bows[verb] = [word for word in positions]

    def is_passive(self):
        verbs = [token for token in self.sentence if token.dep == auxpass]
        if verbs:
            return True
        return False

    def get_action_verbs(self):
        all_verbs = self.get_all_verbs()
        verbs_xcomp = [token for token in all_verbs if (token.dep == xcomp) or (token.dep_ in self.verb_dependencies)]
        self.action_verbs = verbs_xcomp

    def get_words_verb(self, verb):
        return [token for token in verb.children if token.dep in self.dependencies]

    def final_corrections(self):
        correct_characters = multiple_replace(self.corrections_replacements, self.reported)
        correct_thispaper = re.sub(
            '(?:in)? (?:this|their|the|our) (?:present|experimental|numerical)?\s?(?:' + study_terms + ')', '',
            correct_characters, flags=re.I)
        correct_space = re.sub('\s+', ' ', correct_thispaper)
        self.reported = correct_space.replace('and to and', 'and')
        self.reported = reinsert_commas(self.sentence.text, self.reported)

    def add_adverb(self):
        res = ''
        experiment = re.findall('experiment|laboratory', self.sentence.text)
        if experiment:
            res += 'experimentally'

        numeric = re.findall('simulat|numeric', self.sentence.text)
        if numeric:
            if experiment:
                res += ' and numerically'
            else:
                res += 'numerically'
        return res

    def initial_corrections(self):
        sentence = self.sentence.text
        correct_sentence0 = re.sub('(?:\(\d+\)|\[\d+\])', '', sentence)
        correct_sentence = re.sub('(?:\(i+\)|\[i+\])', '', correct_sentence0)
        correct_dir_verbs = re.sub('(?:focus(?:es)? on |(?:to)? understand |deals? with )', 'study ', correct_sentence)
        correct_dir_verbs2 = re.sub('carr(?:y|ied) out ', 'analyzed ', correct_dir_verbs)
        correct_dir_verbs3 = re.sub('aim(?:s|ed)? (?:to|at) ', ' ', correct_dir_verbs2)
        self.sentence = self.nlp(correct_dir_verbs3)

    def generate_for_1_dir_verb(self, verb, try_adverbs=False):
        next_words = [word.text for word in self.bows[verb]]
        adverbs = ''
        if self.is_passive():
            next_words[0] = next_words[0].lower()
        verb_past = verb._.inflect(past_tense)

        if try_adverbs:
            if ('experiment' not in next_words) or ('numeric' not in next_words):
                adverbs = self.add_adverb()

        final = [verb_past] + [adverbs] + next_words
        return final

    def generate(self,review_nb=1,uploaded=False):
        # first corrections
        self.initial_corrections()

        # get bow in the right dependency direction
        self.get_bow()
        self.sort_bows()

        # gather everything
        final = []
        for verb in self.bows:
            final.append(self.generate_for_1_dir_verb(verb))

        for elt in final[1:]:
            elt.insert(0, 'and')
        obj = [elt for bow in final for elt in bow]
        if obj:
            if uploaded:
                link="<b>{}</b><sup>{}</sup>".format(self.authors, review_nb+1)
            else:

                link = self.authors
            # link=self.authors
            final_flatten = [link] + obj
            self.reported = ' '.join(final_flatten) + '.'
            self.final_corrections()




from naimai.utils.transformers import split_list, list2pstring, get_first_char_id, get_last_char_id
import pandas as pd

class NER_BOMR_processor:
    '''
    Process ner BOMR classifier predictions
    '''
    def __init__(self, text, df):
        self.text = text
        self.df = df

    def process_predictionstring(self, elt) -> list:
        '''
        process predictionstring column to split lists into sequences
        :param elt:
        :return:
        '''
        list_pstring = list(map(int, elt.split()))
        list_processed = split_list(list_pstring)
        return list_processed

    def process_df(self):
        '''
        transform original df into df with splitted predictionstring following sequences
        :return:
        '''
        new_dict = {'class': [], 'new_pstring': []}
        self.df['pstring_processed'] = self.df['predictionstring'].apply(self.process_predictionstring)
        for idx in range(len(self.df)):
            elts = self.df.iloc[idx]
            classe = elts['class']
            new_pstrings = elts['pstring_processed']
            if len(new_pstrings) == 1:
                pstring_elt = new_pstrings[0]
                new_dict['class'].append(classe)
                list_pstring = list2pstring(pstring_elt)
                new_dict['new_pstring'].append(list_pstring)
            else:
                for elt in new_pstrings:
                    if elt:
                        new_dict['class'].append(classe)
                        list_pstring = list2pstring(elt)
                        new_dict['new_pstring'].append(list_pstring)
        self.df = pd.DataFrame(new_dict)

    def add_metadata(self):
        '''
        add start, end, start_char and last_char in columns of df
        :return:
        '''
        self.process_df()
        self.df['start'] = self.df['new_pstring'].apply(lambda x: int(x.split()[0]) - 1)
        self.df['end'] = self.df['new_pstring'].apply(lambda x: int(x.split()[-1]))
        self.df['start_char'] = self.df.apply(get_first_char_id, args=(self.text,), axis=1)
        self.df['last_char'] = self.df.apply(get_last_char_id, args=(self.text,), axis=1)
        self.df = self.df.sort_values(by=['start']).reset_index(drop=True)

    def len_intersection_2_ranges(self,rang1: range, rang2: range) -> int:
        '''
        get length of intersection between range1 and range2
        :param rang1: range
        :param rang2:
        :return:
        '''
        xs = set(rang1)
        return len(xs.intersection(rang2))

    def intersection_with_class(self,sentence_range: range, class_ranges: range) -> int:
        '''
        get length of intersection between range of wids on of sentence_range and list of ranges in class ranges
        :param sentence_range:
        :param class_ranges:
        :return:
        '''
        sum = 0
        for rg in class_ranges:
            sum += self.len_intersection_2_ranges(rg, sentence_range)
        return sum

    def get_range_wids_labels(self) -> dict:
        '''
        get, for each label, list of ranges between start word ids and last word ids
        :return:
        '''
        labels = self.df['class'].unique().tolist()
        range_wids_labels = {elt: [] for elt in labels}
        for _, elt in self.df.iterrows():
            class_ = elt['class']
            start, end = elt['start'], elt['end']
            range_wids_labels[class_].append(range(start, end))
        return range_wids_labels

    def get_label_with_most_overlap(self,sentence_range_wids: range, range_wids_labels: dict) -> str:
        '''
        find label with most overlap : with the higher intersection between class ranges and sentence range of wids.
        :return:
        '''
        overlaps = {}
        for cls in range_wids_labels:
            class_ranges = range_wids_labels[cls]
            overlaps[cls]= self.intersection_with_class(sentence_range_wids,class_ranges)
        highest_overlap = max(overlaps.values())
        highest_class = [elt for elt in overlaps if overlaps[elt]==highest_overlap]
        return highest_class[0]



from crossref.restful import Journals
from semanticscholar import SemanticScholar
from tqdm.notebook import tqdm
import time
import random


class ISSN_crawler:
    def __init__(self, issn, field_issn):
        self.issn = issn
        self.field_issn = field_issn
        self.docs = {'title': [], 'authors': [], 'date': [], 'field_paper': [], "abstract": [], "doi": [],
                     "numCitedBy": [], "numCiting": [], "field_issn": []}

    def get_dois(self,idx_start,idx_finish):
        journals = Journals()
        works = journals.works(self.issn)
        for elt in works.select('DOI'):
          try:
            self.docs["doi"].append(elt["DOI"])
          except:
            pass
        if idx_finish==-1:
            self.docs["doi"] = self.docs["doi"][idx_start:]
        else:
            self.docs["doi"] = self.docs["doi"][idx_start:idx_finish]
        print('Len dois : ', len(self.docs['doi']))

    def get_authors(self, authors_list):
        return ', '.join([elt['name'] for elt in authors_list])

    def correct_result(self, result):
        if result:
            return result
        else:
            return ""

    def get_docs(self,idx_start=0, idx_finish=-1,t_min=2, t_max=4,show_tqdm=True,verbose_every=0):
        if not self.docs['doi']:
            self.get_dois(idx_start, idx_finish)
        sch = SemanticScholar(timeout=20)
        dois_to_remove = []
        if show_tqdm:
            range_=tqdm(self.docs['doi'])
        else:
            range_ = self.docs['doi']
        idx=-1
        for doi in range_:
            idx+=1
            try:
                paper = sch.paper(doi)
                slp = random.randint(t_min, t_max)
                time.sleep(slp)

                if paper:
                    if verbose_every:
                        if idx % verbose_every == 0:
                            len_abstracts = len(self.docs['abstract'])
                            print(f"  >> Len docs in {idx} : {len_abstracts}")
                    abstract = self.correct_result(paper["abstract"])
                    if abstract:
                        self.docs['title'].append(self.correct_result(paper["title"]))
                        authors = self.get_authors(paper["authors"])
                        self.docs['authors'].append(self.correct_result(authors))
                        self.docs['date'].append(self.correct_result(paper["year"]))
                        if paper["fieldsOfStudy"]:
                            field = ", ".join(paper["fieldsOfStudy"])
                        else:
                            field = ""
                        self.docs['field_paper'].append(field)
                        self.docs['abstract'].append(abstract)
                        self.docs['numCitedBy'].append(self.correct_result(paper["citationCount"]))
                        self.docs['numCiting'].append(self.correct_result(paper["referenceCount"]))
                        self.docs['field_issn'].append(self.correct_result(self.field_issn))
                    else:
                        dois_to_remove.append(doi)
                else:
                    dois_to_remove.append(doi)
            except:
                dois_to_remove.append(doi)

        for doi in dois_to_remove:
            self.docs['doi'].remove(doi)



'''
Since the arxiv database could not be processed with papers class (in papers/raw.py), many methods are overwritten in
paper_pmc and papers_pmc.

[Documentation here can be more detailed if needed]
'''
import ast
import re
import pandas as pd
from tqdm.notebook import tqdm
from spacy_langdetect import LanguageDetector
from naimai.decorators import update_naimai_dois
import spacy

from naimai.papers.raw import papers, paper_base
from naimai.constants.paths import path_open_citations
from naimai.constants.regex import regex_spaced_chars,regex_keywords
from naimai.constants.nlp import nlp_vocab
from naimai.utils.regex import multiple_replace
from naimai.utils.general import get_soup


class paper_pmc(paper_base):
    def __init__(self ,df ,idx_in_df):
        super().__init__()
        self.database ="pmc"
        self.paper_infos = df.iloc[idx_in_df,:]

    def get_doi(self) -> str:
        self.doi = self.paper_infos['doi']

    def get_fields(self) -> list:
        '''
        get title + journal as fields. use it when you have title + journal
        :return:
        '''
        journal = re.sub('journal', '', self.Journal, flags=re.I)
        self.fields = [self.Title, journal]



    def get_Keywords_from_dict_abstract(self) -> str:
        '''
        get keywords from abstract when it's in a dict format & filter from abstract
        :return:
        '''
        for elt in self.Abstract:
            txt = self.Abstract[elt]
            keywords_in_txt = re.findall(regex_keywords, txt, re.I)
            if keywords_in_txt:
                self.Keywords = keywords_in_txt[0].strip()
                regex_filter = 'key\s?words:\s?' + self.Keywords
                self.Abstract[elt] = re.sub(regex_filter,'',txt,flags=re.I).strip()

    def get_Keywords_from_str_abstract(self)-> str:
        '''
        get keywords from abstract when it's in str format & filter from abstract
        :return:
        '''
        txt = self.Abstract
        keywords_in_txt = re.findall(regex_keywords, txt, re.I)
        if keywords_in_txt:
            self.Keywords = keywords_in_txt[0].strip()
            regex_filter = 'key\s?words:\s?' + self.Keywords
            self.Abstract = re.sub(regex_filter, '', txt, flags=re.I).strip()

    def get_Keywords(self) -> str:
        '''
        Get keywords from abstracts elements & remove them from abstract
        :return:
        '''
        if isinstance(self.Abstract,dict):
            self.get_Keywords_from_dict_abstract()
        elif isinstance(self.Abstract,str):
            self.get_Keywords_from_str_abstract()


    def get_Abstract(self) -> str:
        '''
        clean & stack (if stacked=True) abstract elements into one text & get Keywords from abstract.
        The spaced abstract case (a b s t r a c t..) is considered. if stacked = False, it puts the abstract elements in
        str format. It can returns in dict format if dict_format=True
        :return:
        '''
        abstract_dict_str = self.paper_infos['abstract']
        abstract_dict = ast.literal_eval(abstract_dict_str)
        if 'text' not in abstract_dict.keys():
            abstract_dict = {elt: ' '.join(abstract_dict[elt]) for elt in abstract_dict}

        for elt in abstract_dict:
            clean1=re.sub('abstract', '',abstract_dict[elt], flags=re.I).strip()
            clean2 = self.clean_text(clean1)
            abstract_dict[elt]=clean2

        abstract = ' '.join([elt for elt2 in list(abstract_dict.values()) for elt in elt2])
        no_space = re.findall('\w\w',abstract)
        if no_space: # normal case
            self.Abstract = abstract.replace('-\n', '').replace('\n', ' ')
        else: #spaced abstract, we "despace" 2 times
            despacing1 = re.sub(regex_spaced_chars, r'\1\2', abstract)
            self.Abstract = re.sub(regex_spaced_chars, r'\1\2', despacing1).replace('\n',' ')
        self.get_Keywords()

    def get_Title(self) -> str:
        self.Title = self.paper_infos['title'].replace('-\n', '').replace('\n', ' ')

    def get_Authors(self) -> str:
        authors_list_str = self.paper_infos['authors']
        authors_list = ast.literal_eval(authors_list_str)
        self.Authors = ', '.join(authors_list)

    def get_year(self) -> str:
        self.year = self.paper_infos['year']

    def get_journal(self) -> str:
        self.Journal =  self.paper_infos['journal']

    def get_numCitedBy(self) -> int:
        path = path_open_citations + self.doi
        soup = get_soup(path)
        soup_list = ast.literal_eval(soup.text)
        if isinstance(soup_list,list):
            self.numCitedBy = len(soup_list)


    def replace_abbreviations(self):
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)
            self.Title = multiple_replace(abbreviations_dict, self.Title)

# FIELDS AFTER TITLE JOURNAL
class papers_pmc(papers):
    def __init__(self, papers_path,nlp=None):
        super().__init__() # loading self.naimai_dois & other attributes
        self.data = pd.read_csv(papers_path)
        print('Len data : ', len(self.data))
        if nlp:
            self.nlp = nlp
        else:
            print('Loading nlp vocab..')
            self.nlp = spacy.load(nlp_vocab)
            self.nlp.add_pipe(LanguageDetector(), name='language_detector', last=True)

    def add_paper(self,idx_in_data):
            new_paper = paper_pmc(df=self.data,
                                    idx_in_df=idx_in_data)
            new_paper.get_doi()
            new_paper.get_Title()
            if new_paper.is_paper_english(self.nlp) and isinstance(new_paper.doi,str):
                if not new_paper.is_in_database(self.naimai_dois):
                    new_paper.get_Abstract()
                    new_paper.get_journal()
                    new_paper.get_fields()
                    new_paper.get_Authors()
                    new_paper.get_year()
                    new_paper.replace_abbreviations()
                    try:
                        new_paper.get_numCitedBy()
                    except:
                        pass
                    self.elements[new_paper.doi] = new_paper.save_dict()
                    self.naimai_dois.append(new_paper.doi)


    @update_naimai_dois
    def get_papers(self,update_dois=False,idx_start=0,idx_finish=-1,show_tqdm=False):
        if show_tqdm:
            range_ = tqdm(self.data.iterrows(),total=len(self.data))
        else:
            range_= self.data.iterrows()
        for idx,_ in range_:
            self.add_paper(idx_in_data=idx)


import numpy as np
import pandas as pd
import ast
import re
import random
from tqdm.notebook import tqdm

tqdm.pandas()

from naimai.constants.regex import regex_background, regex_objective,regex_methods,regex_results, regex_objectives
from naimai.papers.full_text.pmc import paper_pmc

def entity_annotate(text, label):
    split = text.split()
    if label=='other':
        entities_labels = ['O' for _ in split]
    else:
        entities_labels = ['B-{}'.format(label)]
        for _ in split[1:]:
            entities_labels.append('I-{}'.format(label))
    return entities_labels

def tostr(lst):
  lst2=[str(elt) for elt in lst]
  return ' '.join(lst2)



class OMRData:
    '''
      Construct data for OMR classification, based on pmc dataframes :
          1. transform structured abstracts to the BOMR structure {Background:.., Objective:.., Methods:.., Results:..}
          2. move objective phrases in background (if exist) to objective section in BOMR
          3. add Other phrases from the body (that would be classified as O in NER labelling), we then get : BOMRO (BOMR + 'Other' section)
          3. convert BOMRO dict to df
          4. Stack & Format the data in the input format for the classifiers
      '''
    def __init__(self, data_df=None,path_data=''):
        if path_data:
            data = pd.read_csv(path_data)
        else:
            data = data_df
        data['is_structured'] = data['abstract'].apply(self.is_structured)
        self.data = data[['doi', 'abstract','body']][data['is_structured'] == True] # only structured abstracts & dois & body are selected
        self.data = self.data.reset_index(drop=True)
        self.transformed_data = None
        self.NER_data = None
        self.BOMRO_class_num = {'ba':0,'ob':1,'me':2,'re':3,'ot':4}

    def is_structured(self,text):
        '''
        check if abstract is structured
        :param text:
        :return:
        '''
        txt2dic = ast.literal_eval(text)
        keys_str = ' '.join(list(txt2dic.keys()))
        if re.findall('text|review|question|meaning|finding',keys_str, flags=re.I):
            return False
        return True

    def is_background(self,text):
        if re.findall(regex_background,text,flags=re.I):
            return True
        return False

    def is_objective(self,text):
        if re.findall(regex_objective,text,flags=re.I):
            return True
        return False

    def is_methods(self,text):
        if re.findall(regex_methods,text,flags=re.I):
            return True
        return False

    def is_results(self,text):
        if re.findall(regex_results,text,flags=re.I):
            return True
        return False

    def abstractKeys2BOMR_format(self,abstract_keys):
        '''
        map abstract keys to BOMR format
        :param abstract_dict:
        :return:
        '''
        BOMR = {'background': [], 'objectives': [], 'methods': [], 'results': []}
        for head in abstract_keys:
            if self.is_background(head):
                BOMR['background'].append(head)
            elif self.is_objective(head):
                BOMR['objectives'].append(head)
            elif self.is_methods(head):
                BOMR['methods'].append(head)
            elif self.is_results(head):
                BOMR['results'].append(head)
        return BOMR

    def correct_BOMR(self, BOMR):
        '''
        objective phrases can often be found in background section. Here we move it to the objective section
        :param BOMR:
        :return:
        '''
        if BOMR['background']:
            obj_phrases = re.findall(regex_objectives,BOMR['background'],flags=re.I)
            if obj_phrases:
                for obj in obj_phrases:
                    BOMR['objectives']+= obj
                    BOMR['background']=BOMR['background'].replace(obj,'')
        BOMR['objectives'] = BOMR['objectives'].replace('\xa0','').strip()
        BOMR['background'] = BOMR['background'].replace('\xa0','').strip()
        BOMR['methods'] = BOMR['methods'].replace('\xa0','').strip()
        BOMR['results'] = BOMR['results'].replace('\xa0','').strip()
        return BOMR

    def abstract2BOMRO(self, abstract_dict):
        '''
        map abstract dict to BOMRO dict
        :param abstract_dict:
        :return:
        '''
        BOMR = {'background': '', 'objectives': '', 'methods': '', 'results': ''}
        if not isinstance(abstract_dict,dict):
            abstract_dict = ast.literal_eval(abstract_dict)
        abstract_keys = list(abstract_dict.keys())
        BOMR_format = self.abstractKeys2BOMR_format(abstract_keys)
        for head in BOMR_format:
            abstract_headers = BOMR_format[head]
            for header in abstract_headers:
                BOMR[head] +=' '.join(abstract_dict[header])
            if self.is_long_sentence(BOMR[head]):
                BOMR[head]=BOMR[head].strip()
            else:
                BOMR[head]=''
        BOMR_corrected = self.correct_BOMR(BOMR) #move obj phrases from background to objective

        idx_df = self.data.index[self.data['abstract']==str(abstract_dict)].to_list()[0]
        other_sentences = self.get_O_entity_label(idx=idx_df)
        if other_sentences:
            BOMR['other'] = [elt.strip() for elt in other_sentences]
            BOMR['other'] = '. '.join(BOMR['other'])
            BOMR['other']=BOMR['other'].strip()
        else:
            BOMR['other']=''

        return BOMR_corrected

    def get_headers_order(self,order='BOMR'):
        '''
        give list of BOMR headers following one of 3 orders :
            1. BOMR
            2. OBMR
            3. MOBR
        the Other section is randomply placed either at the beginning or the end.
        :param order:
        :return: list of headers
        '''
        if order=='BOMR':
            result= ['background', 'objectives', 'methods', 'results']
        elif order=='OBMR':
            result= ['objectives','background','methods','results']
        elif order=='MOBR':
            result= ['methods','objectives','background','results']
        else:
            print('not well written.., returned the default BOMR')
            result= ['background', 'objectives', 'methods', 'results']

        condition=random.choice([0,1])
        if condition:
            result = ['other'] + result
        else:
            result += ['other']
        return result

    def BOMR2dict_data(self,BOMR,id_BOMR,order='BOMR'):
        '''
        transforms BOMR dictionary into data disctionary following the given order. Default order : BOMR. if add_other,
        get other sentences that are not BOMR
        :param BOMR: BOMR dictionary
        :param id_BOMR: id of the BOMR
        :param order: string : BOMR, OBMR or MOBR
        :return:
        '''
        headers= self.get_headers_order(order=order)
        dict_data = {'doi': [], 'text': [], 'start': [], 'end': [], 'predictionstring': [], 'class': [], 'class_num': []}

        id_str_start = 0
        last_word_id = 0
        # id_str_finish = 0
        for head in headers:
            text = BOMR[head]
            if text:
                dict_data['text'].append(text)
                dict_data['start'].append(id_str_start)
                dict_data['end'].append(id_str_start+len(text)-1)
                words_ids = list(np.arange(len(text.split()))+last_word_id+1)
                dict_data['predictionstring'].append(words_ids)
                dict_data['class'].append(head)
                first_class_letter = head[:2]
                dict_data['class_num'].append(self.BOMRO_class_num[first_class_letter])

                id_str_start+=len(BOMR[head])
                last_word_id=words_ids[-1]

        dict_data['doi'] = [id_BOMR]*len(dict_data['text'])

        return dict_data

    def BOMR2text(self,BOMR,order='BOMR'):
        '''
        transforms BOMR dictionary into plain text following the given order. Default order : BOMR (with other at the beginning or end)
        :param BOMR:
        :param order:
        :return:
        '''
        headers = self.get_headers_order(order=order)
        text=''
        for head in headers:
            text +=' ' + BOMR[head]

        return text.strip()

    def BOMR2entities(self,BOMR,order='BOMR'):
        '''
        transforms BOMR dictionary into entities labels : B-label, I-label...
        :param BOMR:
        :return:
        '''
        headers = self.get_headers_order(order=order)
        entities = {}
        for head in headers:
            entities[head] = entity_annotate(text=BOMR[head],label=head)
        return entities

    def data2BOMR_df(self,tqdm=True):
        '''
        convert the data df to BOMR dataframe, with starts, end, etc for each sentence
        :return:
        '''
        if tqdm:
            list_BOMR_dicts = list(self.data['abstract'].progress_apply(self.abstract2BOMRO))
        else:
            list_BOMR_dicts = list(self.data['abstract'].apply(self.abstract2BOMRO))
        list_dois = list(self.data['doi'])
        list_bomr_dfs = []
        for bomr,doi in zip(list_BOMR_dicts,list_dois):
            dict_data = self.BOMR2dict_data(bomr,doi)
            df = pd.DataFrame(dict_data)
            list_bomr_dfs.append(df)
        try:
            dfs_concatenated = pd.concat(list_bomr_dfs, ignore_index=True)
            dfs_concatenated['start']= dfs_concatenated['start'].astype(int)
            dfs_concatenated['end']= dfs_concatenated['end'].astype(int)
            dfs_concatenated['class_num']= dfs_concatenated['class_num'].astype(int)
            dfs_concatenated['predictionstring'] = dfs_concatenated['predictionstring'].apply(tostr)
            self.transformed_data= dfs_concatenated
        except:
            pass

    def data2BOMR_NER_df(self,tqdm=True, weights=(60, 20, 20)):
        '''
        convert the data df to BOMR dataframe only with doi, text and entities. This could be the input to NER model
        :param tqdm: show tqdm
        :param weights: weights of BOMR, OBMR & MOBR
        :return:
        '''
        if tqdm:
            list_BOMR_dicts = list(self.data['abstract'].progress_apply(self.abstract2BOMRO))
        else:
            list_BOMR_dicts = list(self.data['abstract'].apply(self.abstract2BOMRO))
        list_dois = list(self.data['doi'])
        abstract_list = []
        entities_list = []
        order_list=[]
        orders = ['BOMR', 'OBMR', 'MOBR']
        for bomr,doi in zip(list_BOMR_dicts,list_dois):
            order=random.choices(orders, weights=weights, k=1)[0]
            abstract_list.append(self.BOMR2text(bomr,order=order))
            entities=self.BOMR2entities(bomr,order=order)
            entities_vals = []
            for head in entities:
                entities_vals += entities[head]
            entities_list.append(entities_vals)
            order_list.append(order)
        self.NER_data = pd.DataFrame({'doi': list_dois,
                           'text': abstract_list,
                           'entities': entities_list,
                           'order_list': order_list})
    def is_long_sentence(self,sentence: str,threshold=8):
        split = sentence.split()
        if len(split)>threshold:
            return True
        return False


    def get_O_entity_from_text(self,text):
        '''
        get labeled O entities from text
        :param text:
        :return:
        '''
        regex_words_numbers_some = r'[^A-Za-z\s\.,]'
        text_filtered = re.sub(regex_words_numbers_some,'',text)
        text_filtered = re.sub('\s+',' ',text_filtered).strip()
        split = text_filtered.split('.')
        split_filtered = [elt for elt in split if self.is_long_sentence(elt)]
        split_no_obj = [elt for elt in split_filtered if not re.findall(regex_objectives,elt,flags=re.I)]
        if split_no_obj:
            return random.choice(split_no_obj)
        return

    def get_O_entity_label(self,doi=None,idx=None):
        '''
        get labeled O entities from body. Should get either doi or idx
        :param doi:
        :param idx:
        :return:
        '''
        O_labelled_list=[]
        if doi:
            idx=self.data.index[self.data['doi']==doi].to_list()[0]
        else:
            idx=idx
        paper = paper_pmc(df=self.data,idx_in_df=idx)
        paper.get_content()
        if paper.unclassified_section:
            for head in paper.unclassified_section:
                text=paper.unclassified_section[head]
                extracted_other_text= self.get_O_entity_from_text(text)
                if extracted_other_text:
                    O_labelled_list.append(extracted_other_text)
            return list(set(O_labelled_list))
        else:
            return []



from naimai.models.papers_classification.semantic_search import Search_Model
from naimai.utils.regex import transform_field_name
from naimai.constants.paths import aws_root_pdfs, arxiv_pdfs_url, doi_url
import os
import re
import numpy as np
import pandas as pd

class Query_Reviewer:
    def __init__(self, field,encoder=None):
        self.field = field
        self.search_model = Search_Model(field=field, encoder=encoder)

        path_faiss = os.path.join('drive/MyDrive/MyProject/main_pipelines', field, 'encodings.index')
        path_papers = os.path.join('drive/MyDrive/MyProject/main_pipelines', field, 'df_naimai')
        self.search_model.load_naimai_data(path_papers)
        self.search_model.load_faiss_index(path_faiss)


    # def choose_obj_in_mean_lengths_range(self,mean_objs):
    #     mean_lens = 22
    #     if len(mean_objs) == 1:
    #         return mean_objs[0]
    #     else:
    #         # take the closest len to the mean 22
    #         closest_to_mean = min(mean_objs, key=lambda x: abs(len(x.split()) - mean_lens))
    #         return closest_to_mean
    #
    # def choose_obj_beyond_mean_lengths_range(self, references):
    #     mean_lens_radius = np.arange(20, 26)
    #     closest_to_min_radius = min(references, key=lambda x: abs(len(x.split()) - mean_lens_radius[0]))
    #     return closest_to_min_radius
    #
    # def choose_objective(self, references):
    #     mean_lens_radius = np.arange(20, 26)
    #     if len(references) == 1:  # if only one ref, take it
    #         chosen = references[0]
    #         return chosen
    #     if len(references) > 1:  # keep lengths in mean lengths radius 20-25
    #         mean_objs = [elt for elt in references if len(elt.split()) in mean_lens_radius]
    #         if mean_objs:  # we have objs in the lengths radius:
    #             chosen = self.choose_obj_in_mean_lengths_range(mean_objs)
    #             return chosen
    #         else:  # so len(references)>1 and lengths of all refs are either < 20 (so we take the one with max length)
    #             # or > 26 (so we take the one with min length)
    #             chosen = self.choose_obj_beyond_mean_lengths_range(references)
    #             return chosen
    #     return

    def get_year(self,obj):
        year = re.findall('\d+', obj)
        if year:
            return int(year[0])
        return 0

    def get_ref_url(self,obj):
        database = obj['database']
        if database=="mine":
            url = os.path.join(aws_root_pdfs, 'Geophysics', transform_field_name(obj['filename']))
            return url
        if database=="arxiv":
            url = arxiv_pdfs_url + obj['filename']
            return url
        if database=="elsevier":
            url = doi_url + obj['filename']
            return url
        else:
            return '#'

    def obj_formulation(self,obj,prod=False):
        year = self.get_year(obj['objective'])

        if year:
            idx = obj['objective'].index(str(year))
            authors_part = obj['objective'][:idx + 5]
            objective_part = obj['objective'][idx + 5:]
            if prod:
                url = self.get_ref_url(obj)
                authors_part = "<b><a href={}>{}</a></b>".format(url, authors_part)

            return authors_part + objective_part
        return

    def write_by_relevance(self,list_objs,prod=False):
        formulations =[]
        for obj in list_objs:
            formulated = self.obj_formulation(obj,prod=prod)
            if formulated :
                formulations.append(formulated)
        formulations = pd.unique(formulations)
        text = ' '.join(formulations)
        return text

    def write_by_time(self,list_objs,prod=False):
        list_objs = sorted(list_objs, key=lambda x: self.get_year(x['objective']), reverse=True)
        text = self.write_by_relevance(list_objs,prod=prod)
        return text

    def review(self, query,top_n=7,order='Relevance',prod=False):
        objs_query = self.search_model.search(query=query,top_n=top_n)
        list_objs = [{'filename': elt[1]['filename'],
                        'objective': self.choose_objective(elt[1]['reported']),
                        'database': elt[1]['database']} for elt in objs_query]
        if order=='Relevance':
            return self.write_by_relevance(list_objs,prod=prod)
        else:
            return self.write_by_time(list_objs,prod=prod)



import re
from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import random
import requests
import time

def select_span(list_spans):
    for elt in list_spans:
        if 'Posted' in elt.string:
            return elt.string

def find_kwords(soup):
  all_p = soup.find_all(name='p')
  for p in all_p:
    if 'Keywords' in p.getText():
      return p.getText()

class SSRN_Crawler:
    def __init__(self, field, path,t_min=2,t_max=5):
        self.field = field
        self.path = path
        self.soup = {}
        self.total_pages = 999
        self.docs = {'title': [], 'abstract_id': [], 'authors': [], 'date': [], 'field': [], 'nb_page': [],"abstract_box":[], "abstract_text":[],"keywords":[]}
        self.t_min=t_min
        self.t_max=t_max

    def get_soup(self, path):
        header = {}
        header[
            'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        soup = BeautifulSoup(requests.get(path, headers=header, timeout=15).content, 'html.parser')
        slp=random.randint(self.t_min,self.t_max)
        time.sleep(slp)
        return soup

    def get_abstract_box(self,id):
        abstract_path = "https://papers.ssrn.com/sol3/papers.cfm?abstract_id="+id
        abstract_soup = self.get_soup(abstract_path)
        abstract_box = abstract_soup.find_all(name="div", attrs={"class": "box-container box-abstract-main"})
        if not abstract_box:
          #print('soup :  ',abstract_soup)
          print('pproblem with abstract id : ', id)
          #print('abst box : ', abstract_box)
        return abstract_box[0]

    def get_abstract_text(self,abstract_box):
        return abstract_box.find_all(name='div', attrs={"class": "abstract-text"})[0].find('p').getText()

    def get_page_path(self, nb_page):
        split = self.path.split('cfm?')
        page_path = split[0] + 'cfm?'+ 'npage={}&'.format(nb_page) + split[1]
        return page_path

    def get_all_soups(self,first_page=2,last_page=0):
        self.soup[1] = self.get_soup(self.path)
        try:
            self.total_pages = int(self.soup[1].find_all(name="a", attrs={"class": "jeljour_pagination_number"})[-1].string)
            if (last_page>0) and (last_page<self.total_pages):
                self.total_pages = last_page
            for npage in tqdm(range(first_page, self.total_pages + 1)):
                path = self.get_page_path(npage)
                self.soup[npage] = self.get_soup(path)
        except:

            self.total_pages=1
            self.soup[1] = self.get_soup(self.path)

    def get_description_data_npage(self, npage):
        soup = self.soup[npage]
        descriptions = soup.find_all(name="div", attrs={"class": "description"})

        for des in descriptions:
            title = self.get_title(des)
            if title:
                id =  self.get_abstract_id(des)
                if id not in self.docs['abstract_id'] and title not in self.docs['title']:
                  self.docs['title'].append(title)
                  self.docs['abstract_id'].append(id)
                  try:
                      abs_box = self.get_abstract_box(id)
                      self.docs['abstract_box'].append(abs_box)
                  except:
                      print('no abs_box')
                      abs_box=""
                      self.docs['abstract_box'].append(abs_box)

                  if abs_box:
                      try:
                        self.docs['abstract_text'].append(self.get_abstract_text(abs_box))
                      except:
                        self.docs['abstract_text'].append('')
                        print('problem of abstract text in id ', id)
                      try:
                          self.docs['keywords'].append(find_kwords(abs_box))
                      except:
                          self.docs['keywords'].append('')
                  else:
                      self.docs['abstract_text'].append('')
                      self.docs['keywords'].append('')
                  try:
                      self.docs['authors'].append(self.get_authors(des))
                  except:
                      self.docs['authors'].append('')
                  try:
                      self.docs['date'].append(self.get_date(des))
                  except:
                      self.docs['date'].append('')
                  self.docs['field'].append(self.field)
                  self.docs['nb_page'].append(npage)

    def get_title(self, description_soup):
        title= description_soup.find_all(name="a", attrs={"class": "title optClickTitle"})[0].string
        if not title:
            try:
                title = description_soup.find_all(name="a", attrs={"class": "title optClickTitle"})[0].find_all(name="span")[0].string
            except:
                title=''
        return title

    def get_abstract_id(self, description_soup):
        return description_soup.find_all(name="a", attrs={"class": "title optClickTitle"})[0]['href'].split('=')[1]

    def get_date(self, description_soup):
        spans = description_soup.find_all(name="div", attrs={"class": "note note-list"})
        selected_span = [select_span(elt) for elt in spans][0]
        return re.findall('\d+', selected_span)[1]

    def get_authors(self, description_soup):
        authors_elts = description_soup.find_all(name="div", attrs={"class": "authors-list"})
        authors = ''
        for elt in authors_elts:
            authors = ', '.join([elt.string for elt in elt.find_all(name='a')])
        return authors

    def get_docs(self,first_page=2,last_page=0):
        if not self.soup:
            print('Getting soups..')
            self.get_all_soups(first_page=first_page,last_page=last_page)
        print('Getting description..')
        if self.total_pages==1:
            self.get_description_data_npage(self.total_pages)
        for page in tqdm(range(first_page-1, self.total_pages + 1)):
            self.get_description_data_npage(page)




import pickle, gzip, pickletools
import requests
from bs4 import BeautifulSoup
import os
from tqdm.notebook import tqdm
from naimai.constants.paths import naimai_dois_path
import ast
from collections import Counter
import pandas as pd
import re
import time


def load_gzip(path):
    with gzip.open(path, 'rb') as f:
        p = pickle.Unpickler(f)
        return p.load()

def save_gzip(path,obj):
    with gzip.open(path, "wb") as f:
        pickled = pickle.dumps(obj)
        optimized_pickle = pickletools.optimize(pickled)
        f.write(optimized_pickle)

def to_sql(papers,conn,for_semantic=False):
  print('start..')
  paps_df = pd.DataFrame(papers).T
  print('indexing..')
  paps_df['fname'] = paps_df.index
  paps_df = paps_df.reset_index(drop=True)
  print('filling na & sorting ..')
  paps_df=paps_df.fillna('')
  if not for_semantic:
      paps_df['numCitedBy'] = paps_df['numCitedBy'].replace('', 0)
      paps_df = paps_df.sort_values(['numCitedBy'],ascending=False)
  else:
      paps_df = paps_df.sort()
  paps_df = paps_df.astype(str)
  print('saving..')
  paps_df.to_sql(name='all_papers',con=conn)
  print('done!')


def load_gzip_and_update(paths_fnames):
  zips=[]
  if len(paths_fnames)>1:
      for fname in tqdm(paths_fnames):
        zips.append(load_gzip(fname))
      main_dict = zips[0]
      for elt in zips[1:]:
        main_dict.update(elt)
      return main_dict
  all_paps = load_gzip(paths_fnames[0])
  return all_paps

def load_and_combine(path_papers):
    all_files = os.listdir(path_papers)
    paths = [os.path.join(path_papers, fl) for fl in all_files]
    paths = [elt for elt in paths if os.path.isfile(elt) and 'encoding' not in elt and 'sqlite' not in elt]  # keep only files

    all_paps = load_gzip(paths[0])
    for p in paths[1:]:
        paps2 = load_gzip(p)
        all_paps.update(paps2)
    return all_paps

def get_root_fname(fname):
  if '_objectives' in fname:
    return fname
  root_fname= '_'.join(fname.split('_')[:-1])+'_objectives'
  return root_fname

def clean_lst(seq):
    ''' remove duplicates from list '''
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

def get_soup(path):
    header = {}
    header[
        'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
    soup = BeautifulSoup(requests.get(path, headers=header, timeout=120).content, 'html.parser')
    return soup

def listdir_time(path):
  files = os.listdir(path)
  paths_files = [os.path.join(path,elt) for elt in files]
  paths_files.sort(key=lambda x: os.path.getmtime(x))
  files_sorted = [elt.split('/')[-1] for elt in paths_files]
  return files_sorted

def remove_from_naimai_dois(dois_to_remove):
    dois = load_gzip(naimai_dois_path)
    print('len before : ', len(dois))
    for doi in dois_to_remove:
        try:
            dois.remove(doi)
        except:
            pass
    print('len after : ', len(dois))
    save_gzip(naimai_dois_path,dois)

def get_keys_to_filter(cnt):
  keys_to_keep = []
  for k in cnt:
     if cnt[k]>1 and k!='O':
       split = k.split('-')
       if len(split)>1:
         keys_to_keep.append(split[1])
     elif cnt[k]>1 and k=='O':
       keys_to_keep.append(k)

  if keys_to_keep:
    pattern = '|'.join(keys_to_keep)
    keys_to_filter = [elt for elt in cnt if not re.findall(pattern,elt)]
    return keys_to_filter
  else:
    return ''

def correct_ner_data(elt):
  text,entities = elt['text'], ast.literal_eval(elt['entities'])
  cnt = Counter(entities)
  keys_to_filter = get_keys_to_filter(cnt)
  if keys_to_filter:
    new_entities = [elt for elt in entities if elt not in keys_to_filter]
    return new_entities
  return entities

def correct_saved_ner_data(ner_df):
    ner_df['entities']=ner_df.apply(correct_ner_data,axis=1)
    return ner_df

def get_doi_by_title(title,crossref,sleep=2):
  try:
    x = crossref.works(query = title, limit = 1)
    doi = x['message']['items'][0]['DOI']
  except:
    doi=''
  time.sleep(sleep)
  return doi

def ssrn_len_docs(path_csvs,idx_start=0,idx_finish=-1):
  csvs = os.listdir(path_csvs)[idx_start:idx_finish]
  paths = [os.path.join(path_csvs,elt) for elt in csvs]
  dfs = [pd.read_csv(elt) for elt in paths]
  csvs_dict = {field: len(df) for field, df in zip(csvs,dfs)}
  return csvs_dict





from .base import BaseQuerier
from naimai.constants.paths import path_produced
from naimai.utils.general import get_root_fname
from sentence_transformers import SentenceTransformer
import faiss
import os


class SemanticsQuerier(BaseQuerier):
    def __init__(self, field,encoder=None, field_index=None):
        super().__init__(field=field)
        self.field = field
        self.encoder = encoder
        self.field_index = field_index

        self.load_encoder()
        self.load_field_index()


    def load_field_index(self):
        if not self.field_index:
            print('>> Loading field index..')
            path = os.path.join(path_produced, self.field, 'encodings.index')
            self.field_index = faiss.read_index(path)

    def load_encoder(self):
        if not self.encoder:
            print('>> Loading encoder..')
            path = os.path.join(path_produced, self.field, 'search_model')
            self.encoder = SentenceTransformer(path)


    def get_papers_with_bert_semantics(self, query: str, year_from=0, year_to=3000) -> tuple:
        '''
        find papers using encoder & field faiss index.
        '''
        default_top = 60
        encoded_query = self.encoder.encode([query])
        top_n_results = self.field_index.search(encoded_query, default_top)
        ids = top_n_results[1].tolist()[0]
        selected_papers = self.sql_manager.get_by_multiple_ids(ids,year_from=year_from,year_to=year_to)
        selected_papers_fnames = [selected_papers[elt]['fname'] for elt in selected_papers if
                                  selected_papers[elt]['messages']]

        selected_papers_fnames2 = self.remove_duplicated_fnames(selected_papers_fnames)
        selected_papers2 = {elt: selected_papers[elt] for elt in selected_papers_fnames2}
        return selected_papers2, selected_papers_fnames2

    # def find_papers(self, query: str, top_n=5, year_from=0, year_to=3000, verbose=True) -> list:
    #     '''
    #     1. Get query type : simple operator, AND operator or exact match.
    #     2. Get all relevant papers and their fnames
    #     4. Reclassify using tf idf model (need corresponding fname)
    #     5. Rank first papers by numCitedBy using all_papers (need root fname)
    #     6. Get corresponding names
    #     '''
    #
    #     # 1. Get query type : simple operator, AND operator or exact match.
    #     query_type = self.get_query_type(query)
    #     if verbose:
    #         operator = [elt for elt in self.search_operators if self.search_operators[elt]==query_type][0]
    #         print('Operator: ', operator)
    #
    #     # 2. Get all relevant papers and their fnames
    #     if verbose:
    #       print('>> All similar papers selection.. [base.py]')
    #     selected_papers, selected_papers_fnames = self.get_all_similar_papers(query, query_type,year_from=year_from,year_to=year_to,top_n=top_n)
    #
    #     if selected_papers:
    #       root_fnames = [get_root_fname(fname) for fname in selected_papers_fnames]
    #       # root_papers_year_filtered = self.sql_manager.get_by_multiple_fnames(fnames=root_fnames,
    #       #                                                                     year_from=year_from,
    #       #                                                                     year_to=year_to)
    #       # root_fnames_year_filtered = list(root_papers_year_filtered)
    #
    #       # 4. Reclassify using tf idf model (need corresponding fname)
    #       len_query = -1
    #       if query_type!=3:
    #         if verbose:
    #           print('>> Reclassify using tf idf.. [base.py]')
    #         corresponding_papers_fnames = self.get_corresponding_papers(selected_papers_fnames, root_fnames)
    #         corresponding_papers = {fname: selected_papers[fname] for fname in corresponding_papers_fnames}
    #
    #         tf = tfidf_model(query=query, papers=corresponding_papers)
    #         tf_ranked_papers_fnames, scores = tf.get_similar_fnames(top_n=top_n)
    #       else:
    #         # if more than 4 words in query, use tf idf
    #         lemmatized_query = lemmatize_query(self.nlp, query)
    #         len_query = len(lemmatized_query)
    #         if  len_query> 3:
    #           tf = tfidf_model(query=query, papers=selected_papers)
    #           tf.vectorizer.min_df = .05
    #           tf_ranked_papers_fnames, scores = tf.get_similar_fnames(top_n=top_n)
    #         else:
    #           tf_ranked_papers_fnames = selected_papers
    #           scores = [1]*len(tf_ranked_papers_fnames)
    #
    #       if scores[0] < threshold_tf_similarity:
    #         print('>> WARNING : These results may not be relevant!')
    #
    #       # 5. Rank first papers by numCitedBy using all_papers (need root fname)
    #       if verbose:
    #         print('>> numCitedBy Ranking.. [base.py]')
    #       tf_ranked_papers = {fname: selected_papers[fname] for fname in tf_ranked_papers_fnames}
    #       ranked_root_fnames2 = self.rank_papers_with_numCitedBy(tf_ranked_papers,len_query)
    #
    #       # 6. Get corresponding names
    #       corresponding_papers_fnames2 = self.get_corresponding_papers(selected_papers_fnames, ranked_root_fnames2)
    #       if verbose:
    #         print(' ')
    #       return corresponding_papers_fnames2
    #     return []



'''
Attempt to read and extract data from an article in PDF (instead of only the abstract). [Not finished]

[Documentation here can be more detailed if needed]
'''

import re

from naimai.constants.nlp import this_year
from naimai.constants.paths import path_errors_log
from naimai.constants.regex import regex_email, regex_not_converted, regex_references, regex_abstract1, \
    regex_abstract2, regex_words_numbers_some, regex_cid, regex_paper_year
from naimai.constants.replacements import parsing_corrections
from naimai.models.abbreviation import extract_abbreviation_definition_pairs
from naimai.processing import SentenceToProcess
from naimai.utils.pdf2text import convert_pdf_to_txt
from naimai.utils.regex import str1_str2_from_txt, get_pattern, filter_min_length, starts_with_capital, \
    doi_in_text, get_duplicates, multiple_replace
from papers.raw import paper_base


class paper(paper_base):

    def __init__(self, path=''):
        super().__init__()
        self.pdf_path = path
        self.file_name = self.pdf_path.split('/')[-1]
        self.intro_term = 'introduction'
        self.abstract_term = 'abstract'
        self.references_term = 'references'
        self.conclusion_term = 'conclusion'
        self.path_errors_log = path_errors_log
        self.Citations = []
        # self.author_classifier_model=author_classifier_model
        self.converted_text = ''

    def pdf2text(self, use_ocr):
        converted_text = convert_pdf_to_txt(self.pdf_path, is_path=True, use_ocr=use_ocr)
        text_without_cid = re.sub(regex_cid, '', converted_text)
        if text_without_cid:
            self.converted_text = text_without_cid

    def read_pdf(self, use_ocr=False):
        # pdf to raw text
        self.pdf2text(use_ocr)
        if self.converted_text:
            text_only_wordsNnumbers = re.sub(regex_words_numbers_some, '', self.converted_text)
            text_only_wordsNnumbers = re.sub(regex_not_converted, '', text_only_wordsNnumbers)
            duplicated_sentences = get_duplicates(text_only_wordsNnumbers, 5)

            for sentence in duplicated_sentences:
                text_only_wordsNnumbers = text_only_wordsNnumbers.replace(sentence, '')
            text_only_wordsNnumbers = text_only_wordsNnumbers.replace('-\n', '').replace('\n', ' ')

            abbreviations_dic = extract_abbreviation_definition_pairs(doc_text=text_only_wordsNnumbers)
            if abbreviations_dic:
                final_modifs = multiple_replace(abbreviations_dic, text_only_wordsNnumbers)
            else:
                final_modifs = text_only_wordsNnumbers
            self.raw_text = multiple_replace(parsing_corrections, final_modifs)

    def get_Introduction(self, portion=1 / 6):
        """
        assumption : 2.  Introduction is contained in the first portion=1/6 of total lengths of words in paper (1/6 seems to be working in our case : hydraulic)
        :param portion:
        :return:
        """
        headings = self.get_headings()
        if headings:
            headings = self.correct_headings()
            if len(headings) > 1:
                nbs = [int(re.findall('\d', elt)[0]) for elt in headings[:2]]
                if nbs[1] == nbs[0] + 1:
                    self.Introduction = str1_str2_from_txt(headings[0], headings[1], self.raw_text)
            else:
                text_after_intro_term = self.get_txt_after_Introduction()
                self.Introduction = self.get_txt_portion(text_after_intro_term, portion)

        if not self.Introduction:
            text_after_intro_term = self.get_txt_after_Introduction()
            self.Introduction = self.get_txt_portion(text_after_intro_term, portion)

    def get_headings(self):
        headings_pattern = get_pattern(self.raw_text)
        headings = []
        if headings_pattern:
            ptrn = re.compile(headings_pattern, flags=re.M | re.I)
            result = re.findall(ptrn, self.raw_text)
            for elt in result:
                # check lengths
                if (len(elt.split(' ')) < 10) and (len(elt.split(' ')) > 1):
                    headings.append(elt)
            checks = [self.intro_term in elt.lower() for elt in headings]
            if True in checks:
                return headings
        # old method
        return

    def correct_headings(self):
        headings = self.get_headings()
        if headings:
            headings = filter_min_length(headings)
            nbs = [int(re.findall('\d', elt)[0]) for elt in headings]
            one_indx = nbs.index(1)
            nbs = nbs[one_indx:]
            srt_idx = []
            idd = 0
            ref = 99
            for idx, nb in enumerate(nbs):
                if nb == 1 and idx == idd:
                    srt_idx.append(idx)
                    ref = 2
                    idd = idx
                else:
                    if ref == nb:
                        srt_idx.append(idx)
                        ref = nbs[idx] + 1
            return [headings[id] for id in srt_idx]
        return

    def get_txt_after_Introduction(self):
        if self.intro_term in self.raw_text.lower():
            pattern = self.intro_term + '(.*?)\n(.*)'
            resl = re.findall(pattern, self.raw_text, flags=(re.I | re.S))
            if resl:
                text_after_intro_term = resl[0][len(resl[0]) - 1]
            else:
                if self.write_errors:
                    txt_error = 'Little pb in ' + str(self.file_name) + '\n'
                    with open(self.path_errors_log + 'errors_log.txt', 'a') as f:
                        f.write(txt_error)
                    if self.verbose:
                        print(txt_error)
                text_after_intro_term = self.raw_text
        else:
            if self.write_errors:
                txt_error = 'No introduction term in ' + str(self.file_name) + '\n'
                with open(self.path_errors_log + 'errors_log.txt', 'a') as f:
                    f.write(txt_error)
                if self.verbose:
                    print(txt_error)
            text_after_intro_term = self.raw_text
        return text_after_intro_term

    def get_txt_portion(self, text_after_intro_term, portion):
        if text_after_intro_term:
            all_words = text_after_intro_term.split(' ')
            threshold = int(len(all_words) * portion)
            return ' '.join(all_words[:threshold])

    def get_Abstract(self):
        intro_in_txt = self.intro_term in self.raw_text.lower()
        self.get_abtract_term()
        abstract = ''
        if intro_in_txt:
            if self.abstract_term:
                pattern = self.abstract_term + '(.*?)' + self.intro_term
                abstract = self.get_str_from_txt(pattern, idx=0)
            else:
                patrn = '\n(.*?)' + self.intro_term
                abstract = self.get_str_from_txt(patrn, idx=0)

        if not abstract:
            print('no abstract')

        if abstract:
            split = abstract.split('\n\n')
            lst = [len(elt) for elt in split]
            mx = max(lst)
            idx = lst.index(mx)
            abstract = split[idx]
        self.Abstract = abstract

    def get_abtract_term(self):
        self.abstract_term = self.get_str_from_txt(regex_abstract1, flags=(re.M | re.I), idx=0)

        if not self.abstract_term:
            self.abstract_term = self.get_str_from_txt(regex_abstract2, flags=(re.M | re.I), idx=0)

        if not self.abstract_term:
            self.abstract_term = ''

    def get_str_from_txt(self, pattern, flags=(re.I | re.S), idx=-1):
        result = re.findall(pattern, self.raw_text, flags=flags)
        if result:
            return result[idx]
        return

    def get_Conclusion(self, verbose=False):
        try:
            # conclusion term &  next header
            headings = self.get_headings()
            if headings:
                if any(self.conclusion_term in s.lower() for s in headings):
                    conclusion_term = [elt for elt in headings if self.conclusion_term in elt.lower()]
                    if conclusion_term:
                        conclusion_term = conclusion_term[0]
                else:
                    conclusion_term = self.get_str_from_txt('conclusions?')
            else:
                conclusion_term = self.get_str_from_txt('conclusions?')
            next_header = self.header_after_conclusion(headings)

            pattern1 = conclusion_term + '(.*?)' + next_header
            self.Conclusion = self.get_str_from_txt(pattern1, flags=re.S)
            if not self.Conclusion:
                pattern2 = conclusion_term + '(.*)'
                self.Conclusion = self.get_str_from_txt(pattern2, flags=re.S)
        except:
            if verbose:
                print('No conclusion term in ', self.pdf_path)
            pass

    def header_after_conclusion(self, headings):
        next_header = None
        if headings:
            if any(self.conclusion_term in s for s in headings):
                idx_conclusion = [idx for idx, elt in enumerate(headings) if self.conclusion_term in elt][0]
                if len(headings) > idx_conclusion + 1:
                    try:
                        next_header = headings[idx_conclusion + 1]
                        return next_header
                    except:
                        pass

        if not next_header:
            ref_term_in_text = self.get_str_from_txt(self.references_term)
            acknowledgments_in_txt = re.search('acknow?ledge?ments?', self.raw_text, flags=(re.I | re.S))
            if acknowledgments_in_txt:
                next_header = acknowledgments_in_txt.group(0)
            else:
                next_header = ref_term_in_text
            return next_header
        return

    def get_kwords(self):
        kwords_terms = ['key words', 'keywords']
        kwords_term = [elt for elt in kwords_terms if elt in self.raw_text.lower()]
        if kwords_term:
            kwords_term = kwords_term[0]
            idx_kword = self.raw_text.lower().index(kwords_term)

            self.Keywords = self.raw_text[idx_kword:].split('\n\n')[0].lower().replace(kwords_term, '').replace('\n',
                                                                                                                ' ')
        else:
            if self.write_errors:
                txt_error = 'No key words in ' + str(self.file_name) + '\n'
                with open(self.path_errors_log + 'errors_log.txt', 'a') as f:
                    f.write(txt_error)
                if self.verbose:
                    print(txt_error)

    # def get_author_model(self):
    #     if not self.author_classifier_model:
    #         filehandler = open(path_author_classifier,'rb')
    #         self.author_classifier_model = pickle.load(filehandler)

    def get_references(self, verbose=False):
        regex_nb_first = r'^\[?\(?\d+\)?\]?\.?'
        first_list = re.findall(regex_references, self.raw_text, flags=re.S | re.I)
        if first_list:
            first_refs = [elt for elt in first_list[0].split('\n') if elt]
            first_refs = [re.sub(regex_nb_first, '', elt) for elt in first_refs]
            first_refs = [elt for elt in first_refs]

            first_capital = [starts_with_capital(elt) for elt in first_refs]
            for reference, is_capital in zip(first_refs, first_capital):
                if is_capital:
                    self.References.append(reference)
                else:
                    try:
                        self.References[-1] += ' ' + reference
                    except:
                        if verbose:
                            print('no references in ', self.pdf_path)
                        pass

    def get_doi(self):
        doi = doi_in_text(self.raw_text)
        return doi

    def get_emails(self):
        txt_before_intro = self.get_txt_before_Introduction()
        emails = re.findall(regex_email, txt_before_intro, flags=re.I | re.M)
        if emails:
            self.Emails = emails
        else:
            txt_after_intro = self.get_txt_after_Introduction()
            self.Emails = re.findall(regex_email, txt_after_intro, flags=re.I | re.M)

    def get_citations(self):
        if self.Introduction:
            intro = self.Introduction
            stc = SentenceToProcess(intro)
            self.Citations = stc.get_citations()

    def get_txt_before_Introduction(self):
        split = self.raw_text.lower().split(self.intro_term)
        split2_no_empty = []
        if split:
            txt_before_intro = split[0]
            split2 = txt_before_intro.split('\n')
            split2_no_empty = [elt for elt in split2 if elt]
        result = []
        idx_max = 15
        for idx, elt in enumerate(split2_no_empty):
            self.get_abtract_term()
            if (self.abstract_term in elt) or (idx == idx_max):
                break
            result.append(elt)
        return ' '.join(result)

    # def get_list_sentences_between_intro_abstract(self):
    #     split = self.converted_text.split('\n')
    #
    #     for idx, el in enumerate(split):
    #         if self.abstract_term:
    #             found = re.findall(self.abstract_term + '|' + self.intro_term, el, flags=re.I)
    #         else:
    #             found = re.findall(self.intro_term, el, flags=re.I)
    #         if found:
    #             break
    #     split = split[:idx]
    #     split_without_RF = [elt for elt in split if 'River' not in elt]
    #     split3 = [elt.split(';')[0] for elt in split_without_RF if elt]
    #     split4 = []
    #     if split3:
    #         split4 = [elt for elt in clean_authors(split3) if len(elt) > 2 and not re.findall('Engineer|Proce', elt)]
    #     return split4

    # def get_authors(self):
    #     list_sentences = self.get_list_sentences_between_intro_abstract()
    #     GP = self.author_classifier_model
    #     encodings = [self.encoding_model.encode(stc, convert_to_numpy=True) for stc in list_sentences]
    #     authors_list = []
    #     for idx, encod in enumerate(encodings):
    #         GP_pred = GP.predict([encod])
    #         if GP_pred[0]:
    #             authors_list += [list_sentences[idx].replace(' and', '')]
    #     if authors_list:
    #         authors_list = [' '.join(authors_list)]
    #         self.Authors = re.sub(' [a-zA-Z],| [a-zA-Z] | [a-zA-Z]$', '', authors_list[0])

    def get_year(self):
        years = re.findall(regex_paper_year, self.Introduction)
        years_list = [int(elt) for elt in years]
        if years_list:
            year = max(years_list)
            if year < this_year:
                self.year = year
            else:
                self.year = this_year
        else:
            years = re.findall(regex_paper_year, self.raw_text)
            years_list = [int(elt) for elt in years]
            if years_list:
                year = max(years_list)
                if year < this_year:
                    self.year = year
                else:
                    self.year = this_year

# class papers_distil(papers):
#     def __init__(self, all_papers_to_update_dir='', pdfs_dir='', obj_classifier_model=None, author_classifier_model=None,
#                  load_obj_classifier_model=True, load_author_classifier_model=False, load_nlp=True):
#         super().__init__(pdfs_dir=pdfs_dir,obj_classifier_model=obj_classifier_model, author_classifier_model=author_classifier_model,
#                          load_obj_classifier_model=load_obj_classifier_model,
#                          load_author_classifier_model=load_author_classifier_model,
#                          load_nlp=load_nlp)
#         paps_load = self.load(all_papers_to_update_dir)
#         self.all_papers_old = paps_load['elements']
#         self.pdfs_files_names = list(self.all_papers_old.keys())
#
#
#
#     def add_paper(self,pdf_filename,save_dict=True,report=True):
#             pdf_path =self.all_papers_old[pdf_filename]['pdf_path']
#             new_paper = paper(path=pdf_path,
#                               obj_classifier_model=self.obj_classifier_model)
#             new_paper.database='mine'
#             # for abstract, take only the 200 first words
#             try:
#                 new_paper.Abstract =' '.join(self.all_papers_old[pdf_filename]['Abstract'].split()[:200])
#             except:
#                 new_paper.Abstract =' '.join(self.all_papers_old[pdf_filename]['Objective_paper'])
#             new_paper.Authors =self.all_papers_old[pdf_filename]['Authors']
#             new_paper.Publication_year = self.all_papers_old[pdf_filename]['Publication_year']
#             new_paper.Keywords = self.all_papers_old[pdf_filename]['Keywords']
#             # for conclusions, take only one with < 3000 words
#             if len(self.all_papers_old[pdf_filename]['Conclusion'].split()) <3000:
#                 new_paper.Conclusion = self.all_papers_old[pdf_filename]['Conclusion']
#             else:
#                 new_paper.Conclusion = ''
#             new_paper = replace_abbreviations(new_paper)
#             new_paper.get_objective_paper()
#             if report:
#                 new_paper.report_objectives()
#             if save_dict:
#                 self.elements[new_paper.file_name] = new_paper.save_paper_for_training()
#                 self.naimai_elements[new_paper.file_name] = new_paper.save_paper_for_naimai()
#             else:
#                 self.elements[new_paper.file_name] = new_paper
#
#     def get_papers(self,path_chunks='',save_dict=True, report=True):
#         idx=0
#         for pdf_filename in tqdm(self.pdfs_files_names):
#             try:
#                 self.add_paper(pdf_filename,save_dict=save_dict,report=report)
#             except:
#                 pass
#             if idx % 500 == 0 and path_chunks:
#                 print('  Saving idx {} for filename {}'.format(idx, pdf_filename))
#                 self.save(path_chunks)
#             idx += 1
#         print('Objs problem exported in objectives_pbs.txt')



import zipfile
from bs4 import BeautifulSoup
from tqdm.notebook import tqdm
import re

def get_authors(authors_contrib):
  last_name = authors_contrib.find('surname')
  first_name = authors_contrib.find('given-names')
  if last_name:
      return first_name.string + ' ' + last_name.string
  else:
      return ''


class bioRxiv_xml_data:
    def __init__(self):
        self.docs = {'title': [], 'authors': [], 'date': [], 'field_paper': [], "abstract": [], "doi": [],
                     "database": [], "file_name": []}

    def get_doi(self, bs):
        return bs.find_all(name="article-id")[0].string

    def get_topic(self, bs):
        try:
            return bs.find_all(name="subj-group", attrs={'subj-group-type': 'hwp-journal-coll'})[0].find('subject').string
        except:
            return ''

    def get_title(self, bs):
        try :
            return bs.find_all(name="title-group")[0].find('article-title').text
        except:
            return ''

    def get_authors(self, bs):
        try:
            authors_contrib = bs.find_all(name="contrib-group")[0].find_all(name="contrib",
                                                                            attrs={"contrib-type": "author"})
            authors = ', '.join([get_authors(elt) for elt in authors_contrib])
        except:
            authors=""
        return authors

    def get_year(self, bs):
        try :
            return bs.find("year").string
        except:
            return ''

    def get_abstract(self, bs):
        abstract = bs.find("abstract")
        if abstract:
            abstract = abstract.text
            abstract2 = re.sub('abstract', '', abstract, flags=re.I)
            return abstract2.replace('\n', ' ').strip()
        return ''

    def get_kwords(self, bs):
        kwords = bs.find("kwd-group")
        if kwords:
            return ', '.join([elt.string for elt in kwords.find_all('kwd')])
        else:
            return ''

    def get_xml_files(self, archive):
        listfiles = archive.namelist()
        return [elt for elt in listfiles if 'xml' in elt]

    def get_file_data(self, xml_file, xml_file_name):
        bs4_xml = BeautifulSoup(xml_file, "lxml")
        abstract = self.get_abstract(bs4_xml)
        if abstract:
            self.docs['title'].append(self.get_title(bs4_xml))
            self.docs['authors'].append(self.get_authors(bs4_xml))
            self.docs['date'].append(self.get_year(bs4_xml))
            self.docs['field_paper'].append(self.get_topic(bs4_xml))
            self.docs['abstract'].append(abstract)
            self.docs['doi'].append(self.get_doi(bs4_xml))
            self.docs['database'].append("bioRxiv")
            self.docs['file_name'].append(xml_file_name)

    def get_docs(self, path):
        archive = zipfile.ZipFile(path, 'r')
        list_xml_files = self.get_xml_files(archive)

        for fle in tqdm(list_xml_files):
            xml_file = archive.read(fle)
            self.get_file_data(xml_file, fle)
